'''
 Physics Constants
'''
import numpy as np

epsilon0 = 8.8541878176e-12       # vacuum permittivity
mu0      = 4* np.pi*1e-7          # vacuum permiability
c        = 2.99792458e8           # speed of light (m/s)
c_cgs    = c*100.                 # speed of light (cm/s)
q0       = 1.60217662e-19         # electron charge
q0_cgs   = 4.80320427e-10         # electron charge(cgs)
k_B      = 1.380649e-23     # Boltzmann Constant (J/K)
alpha_fine  = 0.0072973525693     # fine structure constant
Da       = 1.66053906660e-27      # atomic mass unit (u or Dalton) (kg)

mass_electron = Da/1822.8884845
mass_hydrogen = Da*1.008 
mass_proton = 1.67262192369e-27



def init_dataset(nsname, datasetnames):
    data = model.datasets.get_child(name = nsname + '_data')
    for n in datasetnames:
           if not data.hasvar(n): 
                data.setvar(n, None)
ans(init_dataset(*args))


'''

   discrete variable integration

     1) integration of discrete varialbe on certain domain/boundary
        in such a way natural to the basis function.
    
        problem dimension (=mesh.Dimension())

        3D domain     H1    volume integration
                      L2    volume integration
                      ND    volume integration of dot prodcuts with coefficients
                      RT    volume integration of dot prodcuts with coefficients

           boundary   H1    area integration
                      L2    (undefined)
                      ND/RT area integration of dot prodcuts with coefficients

        2D domain:    H1/L2 area integration
                      ND/RT area integration of dot prodcuts with coefficients

           boundary:  H1    line integration
                      L2    (undefined)
                      ND    line integration of tangentail component
                      RT    line integration of normal component
        1D domain:    H1    line integration
                      L2    line integration
           boundary:  H1    local value
                      L2    (undefined)


'''
from petram.phys.weakform import get_integrators
from petram.postprocess.pp_model import PostProcessBase
from petram.phys.vtable import VtableElement, Vtable, Vtable_mixin
import numpy as np
import traceback
import warnings

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('Integration(PP)')


bilinintegs = get_integrators('BilinearOps')
linintegs = get_integrators('LinearOps')

data = [("coeff_lambda", VtableElement("coeff_lambda", type='array',
         guilabel="lambda", default='1.0', tip="coefficient",))]


class WeakformIntegrator(PostProcessBase, Vtable_mixin):
    has_2nd_panel = True
    vt_coeff = Vtable(data)

    @property
    def geom_dim(self):  # dim of geometry
        return self.root()['Mesh'].sdim

    @property
    def operator_stock(self):
        if not hasattr(self, "_operator"):
            self._operator = {}
        if not self.reuse_operator:
            return {}
        return self._operator

    def attribute_set(self, v):
        v = super(WeakformIntegrator, self).attribute_set(v)
        v["integration_name"] = 'integ'
        v['coeff_type'] = 'S'
        v['integrator'] = 'MassIntegrator'
        v['variables'] = ''
        v["sdim"] = 2
        v['sel_index'] = ['all']
        v['sel_index_txt'] = 'all'
        v['reuse_operator'] = False
        self.vt_coeff.attribute_set(v)
        return v

    def panel1_param(self):
        import wx
        p = ["coeff. type", "S", 4,
             {"style": wx.CB_READONLY, "choices": ["Scalar", "Vector", "Diagonal", "Matrix"]}]

        names = [x[0] for x in self.itg_choice()]
        names = ['Auto'] + names
        p2 = ["integrator", names[0], 4,
              {"style": wx.CB_READONLY, "choices": names}]

        panels = self.vt_coeff.panel_param(self)
        ll = [["name", "", 0, {}],
              ["variable", self.variables, 0, {}],
              p,
              panels[0],
              p2,
              ["reuse operator", self.reuse_operator, 3, {"text": ""}], ]

        return ll

    def get_panel1_value(self):
        return [self.integration_name,
                self.variables,
                self.coeff_type,
                self.vt_coeff.get_panel_value(self)[0],
                self.integrator,
                self.reuse_operator]

    def import_panel1_value(self, v):
        self.integration_name = str(v[0])
        self.variables = str(v[1])
        self.coeff_type = str(v[2])
        self.vt_coeff.import_panel_value(self, (v[3],))
        self.integrator = str(v[4])
        self.reuse_operator = bool(v[5])

    def panel1_tip(self):
        pass

    def panel2_param(self):
        import wx

        if self.geom_dim == 3:
            choice = ("Volume", "Surface", "Edge")
        elif self.geom_dim == 2:
            choice = ("Surface", "Edge")
        elif self.geom_dim == 1:
            choice = ("Edge", )

        p = ["Type", choice[0], 4,
             {"style": wx.CB_READONLY, "choices": choice}]
        return [p, ["index",  'all',  0,   {'changing_event': True,
                                            'setfocus_event': True}, ]]

    def get_panel2_value(self):
        choice = ["Point", "Edge", "Surface", "Volume", ]
        return choice[self.sdim], self.sel_index_txt

    def import_panel2_value(self, v):
        if str(v[0]) == "Volume":
            self.sdim = 3
        elif str(v[0]) == "Surface":
            self.sdim = 2
        elif str(v[0]) == "Edge":
            self.sdim = 1
        else:
            self.sdim = 1
        self.sel_index_txt = str(v[1])

        from petram.model import convert_sel_txt
        try:
            g = self._global_ns
            arr = convert_sel_txt(self.sel_index_txt, g)
            self.sel_index = arr
        except:
            import traceback
            traceback.print_exc()
            assert False, "failed to convert "+self.sel_index_txt

    def panel2_tip(self):
        pass

    def add_integrator(self, form, engine, integrator, cdim, emesh_idx, isDomain, real):

        self.vt_coeff.preprocess_params(self)
        c = self.vt_coeff.make_value_or_expression(self)

        from petram.helper.variables import var_g
        global_ns = self._global_ns.copy()
        for k in engine.model._variables:
            global_ns[k] = engine.model._variables[k]
        local_ns = {}

        phys = self.root()['Phys'].values()
        for p in phys:
            if p.enabled:
                ind_vars = p.ind_vars
                break
        else:
            assert False, "no phys is enabled"

        from petram.helper.phys_module_util import restricted_integrator

        itg = restricted_integrator(engine, integrator, self.sel_index,
                                    c[0], self.coeff_type[0], cdim,
                                    emesh_idx,
                                    isDomain,
                                    ind_vars, local_ns, global_ns, real)

        if isDomain:
            adder = form.AddDomainIntegrator
        else:
            adder = form.AddBoundaryIntegrator
        adder(itg)


class LinearformIntegrator(WeakformIntegrator):
    def itg_choice(self):
        return linintegs

    @classmethod
    def fancy_menu_name(self):
        return "LinearForm"

    @classmethod
    def fancy_tree_name(self):
        return 'Integration_LF'

    def get_operator(self, engine):
        name = self.variables.strip()
        var1 = engine.model._variables[name]
        emesh_idx1 = var1.get_emesh_idx()
        emesh_idx = emesh_idx1[0]

        fes1 = engine.fespaces[name]
        info1 = engine.get_fes_info(fes1)
        if info1 is None:
            assert False, "fes info is not found"

        if info1['sdim'] == self.sdim:
            isDomain = True
        elif info1['sdim'] == self.sdim+1:
            isDomain = False
        else:
            warnings.warn(
                "Can not perform integration (skipping)", RuntimeWarning)
            return

        isComplex = var1.complex
        new_lf = engine.new_lf

        cdim = 1
        if (info1['element'].startswith('ND') or
                info1['element'].startswith('RT')):
            cdim = info1['sdim']
        else:
            cdim = info1['vdim']

        from petram.helper.phys_module_util import default_lf_integrator
        if self.integrator == 'Auto':
            integrator = default_lf_integrator(info1, isDomain)
        else:
            integrator = self.integrator

        lfr = new_lf(fes1)
        self.add_integrator(lfr, engine, integrator, cdim,
                            emesh_idx, isDomain, True)
        lfr.Assemble()

        if isComplex:
            lfi = new_lf(fes1)
            self.add_integrator(lfi, engine, integrator,
                                cdim, emesh_idx, isDomain, False)
            lfi.Assemble()
        else:
            lfi = None

        from mfem.common.chypre import LF2PyVec
        V = LF2PyVec(lfr, lfi, horizontal=True)

        return V

    def run_postprocess(self, engine):
        dprint1("running postprocess: " + self.name())

        name = self.variables.strip()
        if name not in engine.model._variables:
            assert False, name + " is not defined"

        if engine.ppname_postfix in self.operator_stock:
            V = self.operator_stock[engine.ppname_postfix]
        else:
            V = self.get_operator(engine)
            self.operator_stock[engine.ppname_postfix] = V

        var1 = engine.model._variables[name]
        V1 = engine.variable2vector(var1)

        value = np.array(V.dot(V1), copy=False)

        dprint1("Integrated Value :" + self.integration_name + ":" + str(value))
        engine.store_pp_extra(self.integration_name,
                              value,
                              save_once=True)


class BilinearformIntegrator(WeakformIntegrator):
    def itg_choice(self):
        return bilinintegs

    @classmethod
    def fancy_menu_name(self):
        return "BilinearForm"

    @classmethod
    def fancy_tree_name(self):
        return 'Integration_BF'

    def attribute_set(self, v):
        v = super(BilinearformIntegrator, self).attribute_set(v)
        v["use_conj"] = False
        return v

    def panel1_param(self):
        ll = super(BilinearformIntegrator, self).panel1_param()
        ll[1][0] = "variables"
        ll.append(["use conj (ex. AB^)", self.use_conj, 3, {"text": ""}])
        return ll

    def get_panel1_value(self):
        v = super(BilinearformIntegrator, self).get_panel1_value()
        v.append(self.use_conj)
        return v

    def import_panel1_value(self, v):
        super(BilinearformIntegrator, self).import_panel1_value(v[:-1])
        self.use_conj = bool(v[-1])

    def run_postprocess(self, engine):
        dprint1("running postprocess: " + self.name())

        names = [x.strip() for x in self.variables.split(',')]
        if len(names) != 2:
            assert False, "wrong setting. specify two variables: " + self.variables
        if names[0] not in engine.model._variables:
            assert False, names[0] + " is not defined"
        if names[1] not in engine.model._variables:
            assert False, names[1] + " is not defined"

        var1 = engine.model._variables[names[0]]
        var2 = engine.model._variables[names[1]]

        emesh_idx1 = var1.get_emesh_idx()
        emesh_idx2 = var2.get_emesh_idx()

        if (len(emesh_idx1) != 1 or
            len(emesh_idx2) != 1 or
                emesh_idx1[0] != emesh_idx2[0]):
            assert False, "can not perform integration between different extended-mesh"
        emesh_idx = emesh_idx1[0]

        fes1 = engine.fespaces[names[0]]
        fes2 = engine.fespaces[names[1]]
        info1 = engine.get_fes_info(fes1)
        info2 = engine.get_fes_info(fes2)
        if info1 is None or info2 is None:
            assert False, "fes info is not found"

        if info1['sdim'] == self.sdim:
            isDomain = True
        elif info1['sdim'] == self.sdim+1:
            isDomain = False
        else:
            warnings.warn(
                "Can not perform integration (skipping)", RuntimeWarning)
            return

        isComplex = var1.complex or var2.complex
        if fes1 != fes2:
            new_bf = engine.new_mixed_bf
        else:
            new_bf = engine.new_bf

        cdim = 1
        if (info1['element'].startswith('ND') or
                info1['element'].startswith('RT')):
            cdim = info1['sdim']
        else:
            cdim = info1['vdim']

        from petram.helper.phys_module_util import default_bf_integrator
        if self.integrator == 'Auto':
            integrator = default_bf_integrator(info1, info2, isDomain)
            '''
            if fes1 == fes2:
                if (info1['element'].startswith('ND') or
                    info1['element'].startswith('RT')):
                    integrator = 'VectorFEMassIntegrator'
                elif info1['element'].startswith('DG'):
                    assert False, "auto selection is not supported for GD"
                elif info1['vdim'] > 1:
                    integrator = 'VectorMassIntegrator'
                elif not isDomain:
                    integrator = 'BoundaryMassIntegrator'
                else:
                    integrator = 'MassIntegrator'
            else:
                if (((info1['element'].startswith('ND') or
                      info1['element'].startswith('RT'))and
                     (info2['element'].startswith('ND') or
                      info2['element'].startswith('RT')))):
                    integrator = 'MixedVectorMassIntegrator'
                elif info1['element'].startswith('DG'):
                    assert False, "auto selection is not supported for GD"
                elif (info1['vdim'] == 1 and info1['vdim'] == info2['vdim']):
                    integrator = 'MixedScalarMassIntegrator'                    
                else:
                    assert False, "No proper integrator is found"
            '''
        else:
            integrator = self.integrator

        bfr = new_bf(fes2, fes1)
        self.add_integrator(bfr, engine, integrator, cdim,
                            emesh_idx, isDomain, True)
        bfr.Assemble()

        if isComplex:
            bfi = new_bf(fes2, fes1)
            self.add_integrator(bfi, engine, integrator,
                                cdim, emesh_idx, isDomain, False)
            bfi.Assemble()
        else:
            bfi = None

        from mfem.common.chypre import BF2PyMat

        M = BF2PyMat(bfr, bfi, finalize=True)

        V1 = engine.variable2vector(var1, horizontal=True)
        V2 = engine.variable2vector(var2)
        if self.use_conj:
            try:
                V2 = V2.conj()
            except:
                V2[1] *= -1
        # print(M.shape)
        # print(V1.shape)
        # print(V2.shape)
        value = np.array(V1.dot(M.dot(V2)), copy=False)
        #value = np.array(V1.dot(V2), copy=False)
        dprint1("Integrated Value :" + self.integration_name + ":" + str(value))
        engine.store_pp_extra(self.integration_name,
                              value,
                              save_once=True)



from abc import ABC, abstractmethod
import numpy as np
import warnings
import os
from petram.model import Model
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('Solver')

'''

    Solver configurations


    (GUI)

    SolveStep:   This level defines the block matrix assembled at one time
    Solve:       Solver solves a problem (linear or non-lienar or time-dependent or parametric),
                 using blocks defined in SolveStep

    LinterSolverModel:
                 LinearSolver solves a linear system.
                 Preconditioner is a child LinearSolver of LinearSolver


    (SolverInstance)
n
    SolverInstance: an actual solver logic comes here
       SolverInstance : base class for standard solver
       TimeDependentSolverInstance : an actual solver logic comes here


'''


class SolverBase(Model):
    can_rename = True

    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('phys', self)

    def set_solve_error(self, value):
        self.get_solve_root()._solve_error = value

    def get_solve_root(self):
        '''
        return a model directly under Solver section such as SolveStep
        '''
        obj = self
        solver_root = self.root()['Solver']

        while (not isinstance(obj, SolveStep) and
               obj is not solver_root):
            obj = obj.parent
        return obj

    def eval_text_in_global(self, value, ll=None):
        if not isinstance(value, str):
            return value
        ll = {} if ll is None else ll
        gg = self.root()['General']._global_ns.copy()
        return eval(value, gg, ll)


class SolveStep(SolverBase):
    has_2nd_panel = False

    #
    # GUI and object parameters
    #
    def attribute_set(self, v):
        v['phys_model'] = ''
        v['init_setting'] = ''
        v['postprocess_sol'] = ''
        v['dwc_name'] = ''
        v['use_dwc_pp'] = False
        v['dwc_pp_arg'] = ''
        v['use_geom_gen'] = False
        v['use_mesh_gen'] = False
        v['use_profiler'] = False

        super(SolveStep, self).attribute_set(v)
        return v

    def panel1_param(self):
        ret = [["dwc",   self.dwc_name,   0, {}],
               ["args.",   self.dwc_pp_arg,   0, {}]]
        value = [self.dwc_name, self.dwc_pp_arg]
        return [["Initial value setting",   self.init_setting,   0, {}, ],
                ["Postprocess solution",    self.postprocess_sol,   0, {}, ],
                ["trial phys.", self.phys_model, 0, {}, ],
                [None,  self.use_geom_gen,  3, {
                    "text": "run geometry generator"}],
                [None,  self.use_mesh_gen,  3, {"text": "run mesh generator"}],
                [None,  self.use_profiler,  3, {"text": "use profiler"}],
                [None, [False, value], 27, [{'text': 'use DWC (postprocess)'},
                                            {'elp': ret}]], ]

#                ["initialize solution only",
#                 self.init_only,  3, {"text":""}], ]

    def get_panel1_value(self):
        return (self.init_setting, self.postprocess_sol, self.phys_model,
                self.use_geom_gen,
                self.use_mesh_gen,
                self.use_profiler,
                [self.use_dwc_pp, [self.dwc_name, self.dwc_pp_arg, ]])

    def import_panel1_value(self, v):
        self.init_setting = v[0]
        self.postprocess_sol = v[1]
        self.phys_model = v[2]
        self.use_geom_gen = v[3]
        self.use_mesh_gen = v[4]
        if self.use_geom_gen:
            self.use_mesh_gen = True
        self.use_profiler = bool(v[5])
        self.use_dwc_pp = v[6][0]
        self.dwc_name = v[6][1][0]
        self.dwc_pp_arg = v[6][1][1]

#        self.init_only    = v[2]

    def get_possible_child(self):
        #from solver.solinit_model import SolInit
        from petram.solver.std_solver_model import StdSolver
        from petram.solver.nl_solver_model import NLSolver
        from petram.solver.mg_solver_model import MGSolver
        from petram.solver.ml_solver_model import MultiLvlStationarySolver
        from petram.solver.solver_controls import DWCCall, ForLoop
        from petram.solver.timedomain_solver_model import TimeDomain
        from petram.solver.set_var import SetVar
        from petram.solver.distance_solver import DistanceSolver

        try:
            from petram.solver.std_meshadapt_solver_model import StdMeshAdaptSolver
            return [MultiLvlStationarySolver,
                    TimeDomain,
                    DistanceSolver,
                    StdSolver,
                    StdMeshAdaptSolver,
                    NLSolver,
                    # MGSolver,
                    ForLoop,
                    DWCCall, SetVar]
        except:
            return [MultiLvlStationarySolver,
                    TimeDomain,
                    DistanceSolver,
                    # MGSolver,
                    StdSolver,
                    NLSolver,
                    ForLoop,
                    DWCCall, SetVar]

    def get_possible_child_menu(self):
        #from solver.solinit_model import SolInit
        from petram.solver.std_solver_model import StdSolver
        from petram.solver.nl_solver_model import NLSolver
        from petram.solver.mg_solver_model import MGSolver
        from petram.solver.ml_solver_model import MultiLvlStationarySolver
        from petram.solver.solver_controls import DWCCall, InnerForLoop
        from petram.solver.timedomain_solver_model import TimeDomain
        from petram.solver.set_var import SetVar
        from petram.solver.distance_solver import DistanceSolver

        try:
            from petram.solver.std_meshadapt_solver_model import StdMeshAdaptSolver
            return [("", StdSolver),
                    ("", MultiLvlStationarySolver),
                    ("", NLSolver),
                    ("", TimeDomain),
                    ("extra", DistanceSolver),
                    ("", StdMeshAdaptSolver),
                    ("", InnerForLoop),
                    ("", DWCCall),
                    ("!", SetVar)]
        except:
            return [("", StdSolver),
                    ("", MultiLvlStationarySolver),
                    ("", NLSolver),
                    ("", TimeDomain),
                    ("extra", DistanceSolver),
                    ("", InnerForLoop),
                    ("", DWCCall),
                    ("!", SetVar)]

    @property
    def solve_error(self):
        if hasattr(self, "_solve_error"):
            return self._solve_error
        return (False, "")

    #
    # Data access
    #
    def get_phys(self):
        #
        #  phys for rhs and rows of M
        #
        phys_root = self.root()['Phys']
        ret = []
        for k in self.keys():
            if not self[k].enabled:
                continue
            for x in self[k].get_target_phys():
                if not x in ret:
                    ret.append(x)
            for s in self[k].get_child_solver():
                for x in s.get_target_phys():
                    if not x in ret:
                        ret.append(x)
        return ret

    def get_phys_range(self):
        #
        #  phys for X and col of M
        #
        phys_root = self.root()['Phys']
        phys_test = self.get_phys()
        for n in self.phys_model.split(','):
            n = n.strip()
            p = phys_root.get(n, None)
            if p is None:
                continue
            if not p in phys_test:
                phys_test.append(p)
        return phys_test
        '''
        if self.phys_model.strip() ==  '':
            return self.get_phys()
        else:

            names = [n.strip() for n in names if n.strip() != '']        
            return [phys_root[n] for n in names]
        '''

    def get_target_phys(self):
        return []

    def get_active_solvers(self):
        return [x for x in self.iter_enabled()]

    def get_num_matrix(self, phys_target, set_active_matrix=False, engine=None):
        from petram.engine import max_matrix_num

        num = []
        num_matrix = 0
        active_solves = [self[k] for k in self if self[k].enabled]
        ###

        all_weights = []
        for phys in phys_target:
            for mm in phys.walk():
                if not mm.enabled:
                    continue

                ww = [False]*max_matrix_num
                for s in active_solves:
                    w = s.get_matrix_weight(mm.timestep_config)
                    for i, v in enumerate(w):
                        ww[i] = (ww[i] or v)
                ww = [bool(x) for x in ww]

                mm.set_matrix_weight(ww)
                wt = np.array(ww)
                tmp = int(np.max((wt != 0)*(np.arange(len(wt))+1)))
                num_matrix = max(tmp, num_matrix)

                all_weights.append(mm.get_matrix_weight())

        if set_active_matrix:
            flag = np.sum(all_weights, 0).astype(bool)
            engine.set_active_matrix(flag)
            dprint1("active_matrix flag", flag)
        return num_matrix

    def get_matrix_weight(self, timestep_config):
        raise NotImplementedError(
            "you must specify this method in subclass")

    def is_allphys_real(self):
        phys_target = self.get_phys()
        phys_range = self.get_phys_range()

        phys_real = all([not p.is_complex()
                         for p in phys_target + phys_range])
        return phys_real
    #
    #   data access for init, postprocess, mesh, and goemetry
    #

    def get_init_setting(self):
        names = self.init_setting.split(',')
        names = [n.strip() for n in names if n.strip() != '']
        return [self.root()['InitialValue'][n] for n in names]

    def get_pp_setting(self):
        names = self.postprocess_sol.split(',')
        names = [n.strip() for n in names if n.strip() != '']
        return [self.root()['PostProcess'][n] for n in names]

    def call_run_geom_gen(self, engine):
        name = self.root()['General'].geom_gen
        gen = self.root()['Geometry'][name]
        engine.run_geom_gen(gen)

    def call_run_mesh_gen(self, engine):
        name = self.root()['General'].mesh_gen
        gen = self.root()['Mesh'][name]
        engine.run_mesh_gen(gen)

    def check_and_run_geom_mesh_gens(self, engine):
        flag = False
        if self.use_mesh_gen:
            if self.use_geom_gen:
                self.call_run_geom_gen(engine)
            self.call_run_mesh_gen(engine)
            flag = True
        return flag
    #
    #  verify
    #

    def verify_setting(self):
        try:
            self.get_linearsystem_type_from_modeltree()

        except AssertionError as msg:
            return False, "Can not select linear system type consistently", str(msg)

        return True, "", ""

    def get_linearsystem_type_from_modeltree(self):
        '''
        find appropriate linear system type from model tree. this supports
        two ways to choose linear system type.
           1) LinearSolverModel has an interface to choose the type (old way)
           2) SolverModel has an interface to choose the type (new way)
        '''
        ls_selected = None
        ls_candidates = None

        def make_assertion(cond, message):
            if not cond:
                dprint1("Error: selected ls", ls_selected)
                dprint1("Error: candidate ls", ls_candidates)
                assert cond, message

        def collect_assemble_real(top):
            tmp = []
            for x in top.iter_enabled():
                if hasattr(x, 'assemble_real'):
                    tmp.append(x.assemble_real)
            if True in tmp and False in tmp:
                assert False, "Assemble real is not selected consistently"
            return (True in tmp)
        for x in self.walk():
            if not x.is_enabled():
                continue
            if isinstance(x, LinearSolverModel):
                if x.does_linearsolver_choose_linearsystem_type():
                    assemble_real = collect_assemble_real(self)
                    phys_real = self.is_allphys_real()

                    tmp = x.linear_system_type(assemble_real, phys_real)

                    if ls_selected is None:
                        ls_selected = tmp
                    elif ls_selected == tmp:
                        pass
                    else:
                        make_assertion(
                            False, "Can not select linear system type consistently.(A)")
                else:
                    tmp = x.supported_linear_system_type()
                    if tmp == 'ANY':
                        continue
                    if ls_candidates is None:
                        ls_candidates = set(tmp)
                    else:
                        ls_candidates = ls_candidates & set(tmp)

            if isinstance(x, Solver):
                if x.does_solver_choose_linearsystem_type():
                    tmp = x.get_linearsystem_type_from_solvermodel()
                    if ls_selected is None:
                        ls_selected = tmp
                    elif ls_selected == tmp:
                        pass
                    else:
                        make_assertion(
                            False, "Can not select linear system type consistently. (B)")
                else:
                    pass

        if ls_candidates is not None:
            make_assertion(ls_selected in ls_candidates,
                           "Can not select linear system type consistently. (C)")

        if ls_selected is None:
            make_assertion(
                False, "Model tree does not choose linear system type")
        return ls_selected

    #
    # Preparation for assembly
    #
    def prepare_form_sol_variables(self, engine, n_levels=1):
        solvers = self.get_active_solvers()

        phys_target = self.get_phys()
        phys_range = self.get_phys_range()

        num_matrix = self.get_num_matrix(phys_target,
                                         set_active_matrix=True,
                                         engine=engine)

        engine.set_formblocks(phys_target, phys_range, num_matrix)

        for p in phys_range:
            engine.run_mesh_extension(p)

        engine.run_alloc_sol(phys_range)

#        engine.run_fill_X_block()

    def init(self, engine):
        phys_target = self.get_phys()
        phys_range = self.get_phys_range()

        inits = self.get_init_setting()

        solvers = self.get_active_solvers()
        for solver in solvers:
            inits.extend(solver.get_custom_init())

        engine.run_apply_init(phys_range, inits=inits)
        '''
        if len(inits) == 0:
            # in this case alloate all fespace and initialize all
            # to zero
            engine.run_apply_init(phys_range, 0)
        else:
            for init in inits:
                init.run(engine)
        '''
        # use get_phys to apply essential to all phys in solvestep
        engine.run_apply_essential(phys_target, phys_range)
        engine.run_fill_X_block()

    @debug.use_profiler
    def run(self, engine, is_first=True):
        dprint1("!!!!! Entering SolveStep " + self.name() + " !!!!!")

        wc = self.root()["General"].warning_control
        warnings.simplefilter(wc)
        dprint1("Settiing warning mode :", wc)

        solvers = self.get_active_solvers()

        is_new_mesh = self.check_and_run_geom_mesh_gens(engine)
        if is_first or is_new_mesh:
            engine.preprocess_modeldata()

        # initialize and assemble
        # in run method..
        #   std solver : make sub block matrix and solve
        #   time-domain solver : do step

        # prepare all MG refinement levels here

        flag = True
        lvl = 0
        enabled_flag = engine.model.gather_enebled_flags(engine.model['Phys'])

        while flag:
            engine.level_idx = lvl
            self.prepare_form_sol_variables(engine)
            self.init(engine)

            lvl = lvl + 1
            flag = any([s.create_refined_levels(engine, lvl)
                        for s in solvers])

        engine.model.apply_enebled_flags(engine.model['Phys'], enabled_flag)

        is_first = True
        for solver in solvers:
            is_first = solver.run(engine, is_first=is_first)
            engine.add_FESvariable_to_NS(self.get_phys())
            engine.store_x()
            if self.solve_error[0]:
                dprint1("SolveStep failed " + self.name() +
                        ":" + self.solve_error[1])
                break

        for solver in solvers:
            solver.free_instance()

        postprocess = self.get_pp_setting()
        engine.run_postprocess(postprocess, name=self.name())

        if self.use_dwc_pp:
            engine.call_dwc(self.get_phys_range(),
                            method="postprocess",
                            callername=self.name(),
                            dwcname=self.dwc_name,
                            args=self.dwc_pp_arg)

        warnings.resetwarnings()
        if "PYTHONWARNINGS" in os.environ:
            wc = os.environ["PYTHONWARNINGS"]
            warnings.simplefilter(wc)
        else:
            wc = "Default"
        dprint1("Resettiing warning mode :", wc)
        dprint1("Exiting SolveStep " + self.name())
        return False


class Solver(SolverBase):
    def attribute_set(self, v):
        v['clear_wdir'] = False
        v['init_only'] = False
        v['assemble_real'] = False
        v['save_parmesh'] = False
        v['phys_model'] = ''
        #v['init_setting']   = ''
        v['use_profiler'] = False
        v['probe'] = ''
        v['skip_solve'] = False
        v['load_sol'] = False
        v['sol_file'] = ''
        super(Solver, self).attribute_set(v)
        return v

    @property
    def solve_error(self):
        if hasattr(self, "_solve_error"):
            return self._solve_error
        return (False, "")

    @property
    def instance(self):
        if hasattr(self, "_instance"):
            return self._instance
        return None

    @instance.setter
    def instance(self, value):
        self._instance = value

    def free_instance(self):
        self._instance = None

    def get_phys(self):
        my_solve_step = self.get_solve_root()
        return my_solve_step.get_phys()

    def get_phys_range(self):
        my_solve_step = self.get_solve_root()
        return my_solve_step.get_phys_range()

    def get_target_phys(self):
        names = self.phys_model.split(',')
        names = [n.strip() for n in names if n.strip() != '']
        return [self.root()['Phys'][n] for n in names]

    def get_child_solver(self):
        return []

    def get_custom_init(self):
        return []

    def is_complex(self):
        phys = self.get_phys()
        is_complex = any([p.is_complex() for p in phys])
        if self.assemble_real:
            return False
        # if is_complex: return True
        return is_complex

    def is_converted_from_complex(self):
        phys = self.get_phys()
        is_complex = any([p.is_complex() for p in phys])
        if is_complex and self.assemble_real:
            return True
        return False

    def get_init_setting(self):
        raise NotImplementedError(
            "bug should not need this method")
        '''
        names = self.init_setting.split(',')
        names = [n.strip() for n in names if n.strip() != '']        
        return [self.root()['InitialValue'][n] for n in names]
        '''

    def get_active_solver(self, mm=None, cls=None):
        if cls is None:
            cls = LinearSolverModel
        for x in self.iter_enabled():
            if isinstance(x, cls):
                return x

    def get_active_solvers(self, mm=None, cls=None):
        if cls is None:
            cls = LinearSolverModel

        solvers = []
        for x in self.iter_enabled():
            if isinstance(x, cls):
                solvers.append(x)
        return solvers

    def get_num_matrix(self, phys_target=None):
        raise NotImplementedError(
            "bug should not need this method")

    def get_num_levels(self):
        return 1

    def create_refined_levels(self, engine, lvl):
        '''
        create refined levels and return True if it is created.
        default False (no refined level)
        '''
        return False

    def does_solver_choose_linearsystem_type(self):
        return False

    def get_linearsystem_type_from_solvermodel(self):
        raise NotImplementedError(
            "bug should not need this method")

    @abstractmethod
    def get_matrix_weight(self, *args, **kwargs):
        raise NotImplementedError(
            "bug should not need this method")

    @abstractmethod
    def run(self, engine, is_first=True):
        ...


class SolverInstance(ABC):
    '''
    Solver instance is where the logic of solving a PDF usng
    linearlized matrices (time stepping, adaptation, non-linear...) 
    is written.

    It is not a model object. SolverModel will generate this
    instance to do the actual solve step.
    '''

    def __init__(self, gui, engine):
        self.gui = gui
        self.engine = engine
        self.sol = None
        self.linearsolver_model = None  # LinearSolverModel
        self.linearsolver = None      # Actual LinearSolver
        self.probe = []
        self.linearsolver_model = None

        self._ls_type = self.gui.get_solve_root().get_linearsystem_type_from_modeltree()
        self._phys_real = self.gui.get_solve_root().is_allphys_real()

        if not gui.init_only:
            self.set_linearsolver_model()

    @property
    def ls_type(self):
        return self._ls_type

    @property
    def phys_real(self):
        return self._phys_real

    @ls_type.setter
    def ls_type(self, _v):
        warnings.warn(
            "Setting ls_type does not have any effect.", RuntimeWarning)

    @phys_real.setter
    def phys_real(self, _v):
        warnings.warn(
            "Setting phys_real does not have any effect.", RuntimeWarning)

    def get_phys(self):
        return self.gui.get_phys()

    def get_target_phys(self):
        return self.gui.get_target_phys()

    def get_phys_range(self):
        return self.gui.get_phys_range()

    @property
    def blocks(self):
        return self.engine.assembled_blocks

    def get_init_setting(self):

        names = self.gui.init_setting.split(',')
        names = [n.strip() for n in names if n.strip() != '']

        root = self.engine.model
        return [root['InitialValue'][n] for n in names]

    def set_blk_mask(self):
        # mask defines which FESspace will be solved by
        # a linear solver.
        all_phys = self.get_phys()
        phys_target = self.get_target_phys()
        mask1 = self.engine.get_block_mask(all_phys, phys_target)

        all_phys = self.get_phys_range()
        mask2 = self.engine.get_block_mask(
            all_phys, phys_target, use_range=True)

        self.blk_mask = (mask1, mask2)
        self.engine._matrix_blk_mask = self.blk_mask

    def recover_solution(self, ksol=0):
        '''
        bring linear algebra level solution to gridfunction.
        called when we need a solution vector in gridfucntion vector.
        '''
        engine = self.engine
        phys_target = self.get_phys()
        sol, sol_extra = engine.split_sol_array(self.sol)
        engine.recover_sol(sol)
        extra_data = engine.process_extra(sol_extra)

    def save_solution(self, ksol=0, skip_mesh=False,
                      mesh_only=False, save_parmesh=False,
                      save_mesh_linkdir=None):

        engine = self.engine
        phys_target = self.get_phys()

        if mesh_only:
            engine.save_sol_to_file(phys_target,
                                    mesh_only=True,
                                    save_parmesh=save_parmesh)
        else:
            sol, sol_extra = engine.split_sol_array(self.sol)
            engine.recover_sol(sol)
            extra_data = engine.process_extra(sol_extra)

            engine.save_sol_to_file(phys_target,
                                    skip_mesh=skip_mesh,
                                    mesh_only=False,
                                    save_parmesh=save_parmesh,
                                    save_mesh_linkdir=save_mesh_linkdir)
            engine.save_extra_to_file(extra_data)
        #engine.is_initialzied = False

    def save_probe(self):
        for p in self.probe:
            p.write_file()

    def set_linearsolver_model(self):
        solver = self.gui.get_active_solver()
        if solver is None:
            assert False, "Linear solver is not chosen"

        phys_target = self.get_phys()
        self.linearsolver_model = solver

    def configure_probes(self, probe_txt):
        from petram.sol.probe import Probe

        all_phys = self.get_phys()
        txt = [phys.collect_probes() for phys in all_phys]
        txt = [probe_txt]+txt
        probe_txt = ','.join([t for t in txt if len(t) > 0])

        dprint1("configure probes: "+probe_txt)
        if probe_txt.strip() != '':
            probe_names = [x.strip() for x in probe_txt.split(',')]
            probe_idx = [self.engine.dep_var_offset(n) for n in probe_names]
            for n, i in zip(probe_names, probe_idx):
                self.probe.append(Probe(n, i))

    def allocate_linearsolver(self, is_complex, engine, solver_model=None):
        if solver_model is None:
            solver_model = self.linearsolver_model
        if solver_model.accept_complex:
            linearsolver = self.linearsolver_model.allocate_solver(
                is_complex, engine)
        else:
            linearsolver = solver_model.allocate_solver(
                False, engine)

        return linearsolver

    @abstractmethod
    def solve(self):
        ...

    @abstractmethod
    def compute_rhs(self, M, B, X):
        ...

    @abstractmethod
    def compute_A(self, M, B, X, mask_M, mask_B):
        ...


class TimeDependentSolverInstance(SolverInstance):
    def __init__(self, gui, engine):
        self.st = 0.0
        self.et = 1.0
        self.checkpoint = [0, 0.5, 1.0]
        self._icheckpoint = 0
        self._time = 0.0
        self.child_instance = []
        SolverInstance.__init__(self, gui, engine)

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, value):
        self._time = value
        self.engine.model['General']._global_ns['t'] = value

    @property
    def icheckpoint(self):
        return self._icheckpoint

    @icheckpoint.setter
    def icheckpoint(self, value):
        self._icheckpoint = value

    def set_start(self, st):
        self.st = st
        self.time = st

    def set_end(self, et):
        self.et = et

    @property
    def timestep(self):
        return self._time_step

    def set_timestep(self, time_step):
        self._time_step = time_step

    def set_checkpoint(self, checkpoint):
        self.checkpoint = checkpoint

    def add_child_instance(self, instance):
        self.child_instance.append(instance)

    def solve(self):
        assert False, "time dependent solver does not have solve method. call step"

    @abstractmethod
    def step(self):
        ...


'''
    LinearSolverModel : Model Tree Object for linear solver
    LinearSolver : an interface to actual solver
'''


class LinearSolverModel(SolverBase):
    '''
    Model tree object for a linear solver
    '''
    is_iterative = True

    def get_phys(self):
        return self.parent.get_phys()

    def get_phys_range(self):
        return self.parent.get_phys_range()

    def allocate_solver(self, is_complex=False, engine=None):
        '''
        this method create LinearSolverInstance

        LinearSolverInstance is intermediate obect to prepare LinearSolver
        '''
        raise NotImplementedError(
            "bug. this method sould not be called")

    def prepare_solver(self):
        '''
        this method create LinearSolver. This should return MFEM LinearOperator
        '''
        raise NotImplementedError(
            "bug. this method sould not be called")

    def prepare_solver_with_multtranspose(self):
        '''
        this method create LinearSolver. This should return MFEM LinearOperator
        '''
        raise NotImplementedError(
            "bug. this method sould not be called")

    @abstractmethod
    def real_to_complex(self, solall, M=None):
        ...

    @abstractmethod
    def does_linearsolver_choose_linearsystem_type(self):
        '''
        determins how linearsolvermodel informs the type of linear system to assemble.

        if True: 
            linear_system_type should be implemented.
        if False:
            supported_linear_system_type
        '''

    def linear_system_type(self, assemble_real, phys_real):
        '''
        ls_type: coo  (matrix in coo format : DMUMP or ZMUMPS)
                 coo_real  (matrix in coo format converted from complex 
                            matrix : DMUMPS)
                 blk_interleave (R_fes1, I_fes1, R_fes2, I_fes2,..., I is skipped if real)
                 blk_merged_s (Block operator using ComplexOperator, block_symmetric format)
                 blk_merged   (Block operator using ComplexOperator, asymetric format)

        return None if the model does not specify the linear system type.
        '''
        raise NotImplementedError(
            "bug. this method sould not be called")

    def supported_linear_system_type(self):
        raise NotImplementedError(
            "bug. this method sould not be called")


class LinearSolver(ABC):
    '''
    LinearSolver is an interface to linear solvers such as MUMPS.
    '''
    is_iterative = True

    def __init__(self, gui, engine):
        self.gui = gui
        self.engine = engine
        self._skip_solve = False

    @abstractmethod
    def SetOperator(self, opr, dist=False, name=None):
        ...

    @abstractmethod
    def Mult(self, b, case_base=0):
        ...

    @property
    def skip_solve(self):
        return self._skip_solve

    @skip_solve.setter
    def skip_solve(self, val):
        self._skip_solve = val


def convert_realblocks_to_complex(solall, M, merge_real_imag):
    if merge_real_imag:
        return real_to_complex_merged(solall, M)
    else:
        return real_to_complex_interleaved(solall, M)


def real_to_complex_interleaved(solall, M):
    from petram.mfem_config import use_parallel
    if use_parallel:
        from mpi4py import MPI
        myid = MPI.COMM_WORLD.rank

        of = M.RowOffsets().ToList()
        # of = [np.sum(MPI.COMM_WORLD.allgather(np.int32(o)))
        #      for o in offset]
        # if myid != 0:
        #    return

    else:
        offset = M.RowOffsets()
        of = offset.ToList()

    rows = M.NumRowBlocks()
    s = solall.shape
    nb = rows // 2
    i = 0
    pt = 0
    result = np.zeros((s[0] // 2, s[1]), dtype='complex')
    for j in range(nb):
        l = of[i + 1] - of[i]
        result[pt:pt + l, :] = (solall[of[i]:of[i + 1], :]
                                + 1j * solall[of[i + 1]:of[i + 2], :])
        i = i + 2
        pt = pt + l

    return result


def real_to_complex_merged(solall, M):
    from petram.mfem_config import use_parallel
    if use_parallel:
        from mpi4py import MPI
        myid = MPI.COMM_WORLD.rank

        of = M.RowOffsets().ToList()
        # of = [np.sum(MPI.COMM_WORLD.allgather(np.int32(o)))
        #      for o in offset]
        # if myid != 0:
        #    return

    else:
        offset = M.RowOffsets()
        of = offset.ToList()

    rows = M.NumRowBlocks()
    s = solall.shape
    i = 0
    pt = 0

    result = np.zeros((s[0] // 2, s[1]), dtype='complex')
    for i in range(rows):
        l = of[i + 1] - of[i]
        w = int(l // 2)
        result[pt:pt + w, :] = (solall[of[i]:of[i] + w, :]
                                + 1j * solall[(of[i] + w):of[i + 1], :])
        pt = pt + w
    return result



import time
import numpy as np
import weakref
import traceback
import six
import os
import sys
import tempfile
from six import StringIO
from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD


from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

import multiprocessing as mp
from petram.sol.evaluators import Evaluator, EvaluatorCommon

def data_partition(m, num_proc, myid):
    min_nrows  = m // num_proc
    extra_rows = m % num_proc
    start_row  = min_nrows * myid + (extra_rows if extra_rows < myid else myid)
    end_row    = start_row + min_nrows + (1 if extra_rows > myid else 0)
    nrows   = end_row - start_row
    return start_row, end_row

class BroadCastQueue(object):
   def __init__(self, num):
       self.queue = [None]*num
       self.num = num
       for i in range(num):
           self.queue[i] = mp.JoinableQueue()
           
   def put(self, value, join = False):
       for i in range(self.num):
           self.queue[i].put(value)
       if join:
           for i in range(self.num):
                self.queue[i].join()
                
   def put_single(self, value, join=False):
       self.queue[0].put(value)
       if join:
          self.queue[0].join()
   def join(self):         
       for i in range(self.num):
           self.queue[i].join()
           
   def close(self):         
       for i in range(self.num):
           self.queue[i].close()

   def __getitem__(self, idx):
       return self.queue[idx]

   
class EvaluatorMPChild(EvaluatorCommon, mp.Process):
    def __init__(self, task_queue, result_queue, myid, rank,
                 logfile = False, text_queue = None):

        mp.Process.__init__(self)
        EvaluatorCommon.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue
        self.text_queue = text_queue        
        self.myid = myid
        self.rank = rank
        self.agents = {}
        self.logfile = logfile
        print("log file")
        #self.logfile = 'log'
        self.use_stringio = False
        self.solfiles = None
        
        ## enable it when checking performance        
        self.use_profiler = False

    def run(self, *args, **kargs):
        # this can not be in init, sicne it is not pickable
        self.solvars = WKD()
        
        if self.logfile == 'suppress':
            sys.stdout = open(os.devnull, 'w')
        elif self.logfile == 'queue':
            self.use_stringio = True
        elif self.logfile == 'log':
            path = os.path.expanduser('~/MPChild'+ str(os.getpid()) + '.out')
            sys.stdout = open(path, "w")
        else:
            pass
        while True:
            time.sleep(0.01)
            try:
               task = self.task_queue.get(True)
            except EOFError:
                self.result_queue.put((-1, None))                
                self.task_queue.task_done()
                continue

            if task[0] == -1:
                self.task_queue.task_done()
                break
            try:
                if self.use_stringio:
                    stringio = StringIO()
                    org_sys_stdout = sys.stdout
                    org_sys_stderr = sys.stderr
                    sys.stdout = stringio
                    sys.stderr = stringio
                    
                if self.use_profiler:
                    import cProfile
                    pr = cProfile.Profile()
                    pr.enable()

                #print("got task", task[0], self.myid, self.rank)
                if task[0] == 1: # (1, cls, param) = make_agents
                    cls = task[1]
                    params = task[2]
                    kwargs = task[3]

                    if self.solfiles is None and cls != 'Probe':
                        continue
                    self.make_agents(cls, params, **kwargs)

                elif task[0] == 2: # (2, solfiles) = set_solfiles
                    self.set_solfiles(task[1])
                    ll = len(self.solfiles) if self.solfiles is not None else 0
                    value = (self.myid, ll, None)

                elif task[0] == 3: # (3, mfem_model) = set_model
                    self.set_model(task[1])

                elif task[0] == 4: # (4,)  = load_solfiles
                    if self.solfiles is None: continue                    
                    self.load_solfiles()

                elif task[0] == 5: # (5,)  = phys_path
                    self.phys_path = task[1]

                elif task[0] == 6: # (6, attr)  = process_geom
                    if self.solfiles is None: continue
                    self.call_preprocesss_geometry(task[1], **task[2])

                elif task[0] == 7: # (7, expr)  = eval
                    if self.solfiles is None:
                        value = (self.myid, None, None)
                    else:
                        value =  self.eval(task[1], **task[2])

                elif task[0] == 8: # (8, expr)  = eval_probe
                    value = self.eval_probe(task[1], task[2], task[3])
                    value = (self.myid, value[0], value[1])

                elif task[0] == 9: # (8, expr)  = make_probe_agents
                    cls = task[1]
                    params = task[2]
                    kwargs = task[3]
                    self.make_probe_agents(cls, params, **kwargs)
                    
                elif task[0] == 10: # (10, expr)  = eval_pointcloud
                    value = self.eval_pointcloud(task[1], **task[2])
                    
                elif task[0] == 11: # (11, expr)  = eval_integral
                    value = self.eval_integral(task[1], **task[2])

            except:
                traceback.print_exc()
                err = traceback.format_exc()
                value = (self.myid, None, None)
            finally:
                self.task_queue.task_done()

                if self.use_profiler:
                    import  pstats                    
                    pr.disable()
                    ps = pstats.Stats(pr,
                                      stream=sys.stdout).sort_stats('cumulative')
                    ps.print_stats()


                if task[0] in [2, 7, 8, 10, 11]:
                    self.result_queue.put(value)
                    
                if self.use_stringio:
                    output = stringio.getvalue()
                    stringio.getvalue()
                    sys.stdout = org_sys_stdout
                    sys.stderr = org_sys_stderr 
                    stringio.close()
                    self.text_queue.put(output)
                    
        #end of while
        self.task_queue.close()
        self.result_queue.close()
        
    def set_solfiles(self, solfiles):
        st, et = data_partition(len(solfiles.set), self.rank, self.myid)
        s = solfiles[st:et]
        print(s)
        if len(s) > 0:
            self.solfiles_real = s
            self.solfiles = s
        else:
            self.solfiles = None

    def set_model(self, model_path):
        try:
            from petram.engine import SerialEngine
            s = SerialEngine(modelfile = model_path)

            if (os.path.basename(model_path) == 'model_proc.pmfm'):
                s.build_ns()
            else:
                s.run_config()            
                model_path = os.path.join(os.path.dirname(model_path), 'model_proc.pmfm')
                if self.myid == 0:
                    s.model.save_to_file(model_path,
                                         meshfile_relativepath=False)

            self.model_real = s.model
        except:
            print(traceback.format_exc())
            return self.myid, None, traceback.format_exc()
            
        super(EvaluatorMPChild, self).set_model(s.model)

    def call_preprocesss_geometry(self, attr, **kwargs):
        solvars = self.load_solfiles()
        for key in six.iterkeys(self.agents):
            evaluators = self.agents[key]
            for o in evaluators:
                o.preprocess_geometry([key], **kwargs)                
        
    def eval(self, expr, **kwargs):
        phys_path = self.phys_path
        phys = self.mfem_model()[phys_path]
        solvars = self.load_solfiles()
        
        if solvars is None: return self.myid, None, None

        data = []
        attrs = []
        for key in six.iterkeys(self.agents): # scan over battr
            data.append([])
            attrs.append(key)                                  
            evaluators = self.agents[key]
            for o, solvar in zip(evaluators, solvars): # scan over sol files
                try:
                     v, c, a = o.eval(expr, solvar, phys, **kwargs)
                except:
                     import traceback
                     return self.myid, None, traceback.format_exc()
                     
                if v is None:
                    v = None; c = None; a = None
                data[-1].append((v, c, a))

        return self.myid, data, attrs

    def eval_pointcloud(self, expr, **kwargs):
        if self.phys_path == '': return None, None, None

        #use_pr = True
        #import cProfile
        #if use_pr:
        #    pr = cProfile.Profile()
        #    pr.enable()
        
        phys = self.mfem_model()[self.phys_path]
        solvars = self.load_solfiles()

        if solvars is None: return self.myid, None, None        

        export_type = kwargs.get('export_type', 1)
        
        data = []
        attrs = []
        offset = 0

        key = list(self.agents)[0]
        
        vdata = [] # vertex
        cdata = [] # data
        adata = [] # array idx     
        attrs.append(key)                                  
        evaluators = self.agents[key]
        
        for o, solvar in zip(evaluators, solvars): # scan over sol files
           try:
               v, c, a = o.eval(expr, solvar, phys, **kwargs)
           except:
               import traceback
               return self.myid, None, traceback.format_exc()
               
           if v is None:
               v = None; c = None; a = None
           else:
               vdata.append(v)
               cdata.append(c)
               adata.append(a)

        if len(adata)==0: return None, None, None

        ptx  = vdata[0]
        ddim = cdata[0].shape[1:]  # data dim

        shape = adata[0].shape
        shape_d = sum((shape, ddim), ())
                      
        attrs = np.zeros(shape, dtype=int)-1

        data  = np.zeros(shape_d, dtype=cdata[0].dtype)

        #print("data shape", data.shape)
                
        for v, c, a in zip(vdata, cdata, adata):
            idx = (a != -1)
            if np.sum(idx) == 0: continue
            attrs[idx] = a[idx]
            data[idx] = c

        #if use_pr:
        #    path = os.path.expanduser('~/MPChild_profile'+ str(os.getpid()) + '.out')
        #    pr.dump_stats(path)
        #    pr.disable()
        return ptx, data, attrs

    def eval_integral(self, expr, **kwargs):
        phys_path = self.phys_path
        phys = self.mfem_model()[phys_path]
        solvars = self.load_solfiles()
        
        if solvars is None:
            return self.myid, None, None

        data = 0.0
        for key in six.iterkeys(self.agents): # scan over battr
            evaluators = self.agents[key]
            for o, solvar in zip(evaluators, solvars): # scan over sol files
                try:
                     v = o.eval_integral(expr, solvar, phys, **kwargs)
                except:
                     import traceback
                     return self.myid, None, traceback.format_exc()
                     
                if v is None:
                    v = None
                data = data + v

        return self.myid, data, 0
        
    def eval_probe(self, expr, xexpr, probes):
        if self.phys_path == '': return None, None
        
        phys = self.mfem_model()[self.phys_path]
        evaluator = self.agents[1][0]
        
        return evaluator.eval_probe(expr, xexpr, probes, phys)
        

class EvaluatorMP(Evaluator):
    def __init__(self, nproc = 2, logfile = False):
        super(EvaluatorMP, self).__init__()
        print("new evaluator MP", nproc)
        self.init_done = False        
        self.tasks = BroadCastQueue(nproc)
        self.results= mp.JoinableQueue() 
        self.workers = [None]*nproc
        self.solfiles = None
        self.failed = False
        self.closed = False
        
        if logfile == 'queue':
            self.text_queue = mp.Queue()
        else:
            self.text_queue = None

        for i in range(nproc):
            w = EvaluatorMPChild(self.tasks[i], self.results, i, nproc,
                                 logfile = logfile,
                                 text_queue = self.text_queue)
            self.workers[i] = w
            time.sleep(0.1)
        for w in self.workers:
            w.daemon = True
            w.start()
        
    def __del__(self):
        if not self.closed:
            self.terminate_all()

    def set_model(self, model):
        #print("looking for model file in " + os.getcwd())
        #print(model)
        #print(model.local_sol_path)
        
        tmpdir = model.local_sol_path
        file1 = os.path.join(tmpdir, "model_proc.pmfm")
        file2 = os.path.join(tmpdir, "model.pmfm")
        
        if os.path.exists(file1):
            self.tasks.put((3, file1), join = True)
        elif os.path.exists(file2):
            self.tasks.put((3, file2), join = True)
        else:
            assert False, "No model file in " + os.getcwd()

        self._mfem_model_bk = model

        #import tempfile, shutil
        #tmpdir = tempfile.mkdtemp()
        #model_path = os.path.join(tmpdir, 'model.pmfm')
        #model.save_to_file(model_path,
        #                   meshfile_relativepath = False)
        #self.tasks.put((3, model_path), join = True)

        #shutil.rmtree(tmpdir)
        
    def set_solfiles(self, solfiles):
        self.solfiles = solfiles
        self.tasks.put((2, solfiles))
        res = [self.results.get() for x in range(len(self.workers))]
        for x in range(len(self.workers)):
            self.results.task_done()

        isnotNone = []
        for v, c, a in res: # handle (myid, error, message)
            if c is None and v is not None:
                continue
            isnotNone.append(v)
        if len(isnotNone) == 0:
            assert False, "solution may not exist"

    def make_agents(self, name, params, **kwargs):
        super(EvaluatorMP, self).make_agents(name, params, **kwargs)
        self.tasks.put((1, name, params, kwargs))

    def load_solfiles(self, mfem_mode = None):
        self.tasks.put((4, ), join = True)

    def set_phys_path(self, phys_path):
        self.tasks.put((5, phys_path))
        
    def validate_evaluator(self, name, attr, solfiles, isFirst=False, **kwargs):
        redo_geom = False
        #if  self.solfiles is None or self.solfiles() is not solfiles:
        if (self.solfiles is None or
            self.solfiles.is_different_timestamps(solfiles)):

            if self.solfiles is None:
                print("self.solfiles is None")
            else:
                print("new file time stamp")
                
            print("new solfiles (reloading model from)")
            self.set_model(self._mfem_model_bk)
            self.set_solfiles(solfiles)
            self.load_solfiles()
            redo_geom = True

        else:
            print("same solfiles")            
        if not super(EvaluatorMP, self).validate_evaluator(name, attr, **kwargs):
            redo_geom = True
        if not self.init_done: redo_geom = True
        if not redo_geom: return
        
        if isFirst:
            self.init_done = True            
            return

        self.make_agents(self._agent_params[0],
                         attr, **kwargs)
        #self.tasks.put((6, attr, kwargs))        
        self.init_done = True
        
    def eval(self, expr, merge_flag1, merge_flag2, **kwargs):
        self.tasks.put((7, expr, kwargs), join = True)

        res = [self.results.get() for x in range(len(self.workers))]
        for x in range(len(self.workers)):
            self.results.task_done()

        for v, c, a in res: # handle (myid, error, message)
            if c is None and v is not None and a is not None:
                assert False, a
            
        # sort it so that the data is in the order of workes.rank
        res = [(x[1], x[2]) for x in sorted(res)]

        for x1, x2 in res:
            if x1 is None and x2 is not None:
                assert False, x2
            
        results = [x[0] for x in res if x[0] is not None]
        attrs = [x[1] for x in res if x[0] is not None]
        attrs = attrs[0]

        data = [None]*len(attrs)

        for kk, x in enumerate(results):
            for k, y in enumerate(x):
                if data[k] is None: data[k] = y
                else: data[k].extend(y)
        num_files = len(data[0])
        def omit_none(l):
            return [x for x in l if x is not None]

        if merge_flag1:
            #data0 = [None]*len(attrs)
            data0 = []
            offset = 0            
            for k, x in enumerate(data):
                if merge_flag2: offset = 0                
                vdata, cdata, adata = zip(*x)
                if len(omit_none(vdata)) == 0: continue
                for c, a in zip(cdata, adata):
                    if c is not None:                    
                        a += offset
                        offset = len(c) + offset
                data0.append([(np.vstack(omit_none(vdata)),
                              np.hstack(omit_none(cdata)),
                              np.vstack(omit_none(adata)))])
            data = data0

        # eliminate non-existent attribute
        data0 = []; attrs0 = []
        for x, a in zip(data, attrs):
            if x is not None:
                data0.append(x)
                attrs0.append(a)
        data = data0; attrs = attrs0                
        
        if merge_flag1 and not merge_flag2:
            x0 = [x[0][0] for x in data]
            if len(x0) == 0:
                assert False, "No slice data point"
            vdata = np.vstack(x0)
            cdata = np.hstack([x[0][1] for x in data])
            adata = np.vstack([x[0][2] for x in data])
            data = [(vdata, cdata, adata)]
        elif merge_flag1:
            data0 = []
            for x in data: data0.extend(x)
            data = data0
        elif not merge_flag2:
            keys = attrs
            data0 = []
            attr = []
            for idx in range(num_files): # for each file
                vdata = []
                cdata = []
                adata = []
                offset = 0
                for idx0, key in enumerate(keys):
                    d1 = data[idx0][idx]
                    if d1[0] is None: continue
                    vdata.append(d1[0])
                    cdata.append(d1[1])
                    adata.append(d1[2]+offset)
                    offset = offset + d1[1].shape[0]
                if offset == 0: continue
                dd  = (np.vstack(vdata), np.hstack(cdata), np.vstack(adata))
                data0.append(dd)
                attr.append(key)
            attrs = list(set(attr))
            data = data0
        else:
            data0 = []
            for x in data:
                data0.extend([xx for xx in x if xx[0] is not None])
            data = data0
        return data, attrs

    def eval_pointcloud(self, expr, **kwargs):
        self.tasks.put((10, expr, kwargs), join = True)

        res = [self.results.get() for x in range(len(self.workers))]
        for x in range(len(self.workers)):
            self.results.task_done()

        res = [x for x in res if x[-1] is not None]

        if len(res) == 0:
            return None, None, None
        
        for v, c, a in res: # handle (myid, error, message)
            if c is None and v is not None and a is not None:
                assert False, a

        ptx, data, attrs = res[0]

        for v, c, a in res[1:]:
            idx = (a != -1)
            if np.sum(idx) == 0: continue
            #print(a)
            #print(attrs)
            attrs[idx] = a[idx]
	    #print(data.shape, c.shape)                                                                                                                         
            data[idx] = c[idx]

        return ptx, data, attrs

    def eval_integral(self, expr, **kwargs):
        self.tasks.put((11, expr, kwargs), join = True)

        res = [self.results.get() for x in range(len(self.workers))]
        for x in range(len(self.workers)):
            self.results.task_done()

        res = [x for x in res if x[-1] is not None]

        if len(res) == 0:
            return None

        v = 0
        for _myid, vv, _extra in res: # handle (myid, error, message)
            if vv is not None:
                v = v + vv
            else:
                assert False, _extra

        return v
            
    def eval_probe(self, expr, xexpr, probes):
        self.tasks.put_single((8, expr, xexpr, probes), join = True)
        res = self.results.get()# for x in range(len(self.workers))]

        self.results.task_done()
        return res

    def make_probe_agents(self, name, params, **kwargs):
        print("make_probe_agents")
        super(EvaluatorMP, self).make_agents(name, params, **kwargs)
        self.tasks.put((9, name, params, kwargs))
    
    def terminate_all(self):
        #print('terminating all')      
        #num_alive = 0
        #for w in self.workers:
        #    if w.is_alive(): num_alive = num_alive + 1
        #for x in range(num_alive):
        if self.closed: return
        self.tasks.put([-1])
        self.tasks.join()
        
        self.tasks.close()
        self.results.close()
        self.results.cancel_join_thread()
        
        if self.text_queue is not None:
            self.text_queue.close()
            self.text_queue.cancel_join_thread()

        self.closed = True
        for w in self.workers:
            w.join()
        
        print('joined')
        
    def terminate_allnow(self):
        return self.terminate_all()

    




from __future__ import print_function

class CommWorld(object):
    def __init__(self):
        self.size = 1
        self.rank = 0
        
    def Barrier(self):
        pass
    def bcast(self, *args, **kwargs):
        return args[0]

    def gather(self, *args, **kwargs):
        return [args[0]]
    
    def allgather(self, *args, **kwargs):
        return [args[0]]
    
class MPIclass(object):
    def __init__(self, *args, **kwargs):
        self.COMM_WORLD = CommWorld()
MPI = MPIclass()

def nicePrint(x):
    print(x)

        



from .solver_model import Solver
import numpy as np

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('MUMPSModel')

class AMS(Solver):
    has_2nd_panel = False
    accept_complex = False
    def init_solver(self):
        pass
    
    def panel1_param(self):
        return [["log_level",   self.log_level,  400, {}],
                ["max  iter.",  self.maxiter,  300, {}],
                ["rel. tol",    self.reltol,  300,  {}],
                ["abs. tol.",   self.abstol,  300, {}],
                ["restart(kdim)", self.kdim,     400, {}],
                ["preconditioner", self.preconditioner,     0, {}],]    
    
    def get_panel1_value(self):
        return (long(self.log_level), self.maxiter,
                self.reltol, self.abstol, self.kdim,
                self.preconditioner)
    
    def import_panel1_value(self, v):
        self.log_level = long(v[0])
        self.maxiter = v[1]
        self.reltol = v[2]
        self.abstol = v[3]
        self.kdim = v[4]        
        self.preconditioner = v[5]
        
    def attribute_set(self, v):
        v = super(AMS, self).attribute_set(v)
        v['log_level'] = 0
        v['maxiter'] = 200
        v['reltol']  = 1e-7
        v['abstol'] = 1e-7
        v['kdim'] =   50
        v['preconditioner'] = 'AMS'
        return v
    
    def verify_setting(self):
        if not self.parent.assemble_real:
            root = self.root
            phys = root['Phys'][self.parent.phys_model]
            if phys.is_complex: return False, "Complex Problem not supported.", "AMS does not support complex problem"
        return True, "", ""



import traceback
from petram.model import Model
from petram.phys.vtable import VtableElement, Vtable, Vtable_mixin

data = [("init_value", VtableElement("init_value", type='array',
                                     guilabel="expression", default=0.0, tip="expression",))]


class InitSetting(Model, Vtable_mixin):
    can_rename = True
    has_2nd_panel = False
    vt_coeff = Vtable(data)

    @property
    def _global_ns(self):
        # used for text box validator
        return self.root()['General']._global_ns

    def attribute_set(self, v):
        v = super(InitSetting, self).attribute_set(v)
        v["phys_model"] = ''
        v["init_var"] = ''
        v["init_mode"] = 0
        #v["init_value_txt"]    = '0.0'
        v["init_path"] = ''
        v["init_dwc_name_params"] = ('', '')
        self.vt_coeff.attribute_set(v)
        return v

    def panel1_param(self):
        from petram.pi.widget_init import InitSettingPanel
        return [["physics model",   self.phys_model,  0, {}, ],
                ["variable name (\"\"=all)",   self.init_var,  0, {}, ],
                [None, None, 99, {'UI': InitSettingPanel, 'validator': self.check_phys_expr_array}], ]

    def _init_eval(self, value):
        gg = self.root()['General']._global_ns.copy()

        from petram.helper.variables import var_g
        ll = var_g.copy()

        return eval(value, gg, ll)

    # def init_validator(self, value, param, ctrl):
    #    try:
    #        x = self._init_eval(value)
    #    except:
    #        import traceback
    #        traceback.print_exc()
    #        return False
    #    return True

    def get_panel1_value(self):
        init_value_txt = self.vt_coeff.get_panel_value(self)[0]
        return [self.phys_model, self.init_var,
                (self.init_mode, init_value_txt, self.init_path, self.init_dwc_name_params), ]

    def import_panel1_value(self, v):
        self.phys_model = str(v[0])
        self.init_var = str(v[1])
        self.init_mode = v[2][0]
        #self.init_value_txt = v[1][1]
        self.init_path = v[2][2]
        self.init_dwc_name_params = v[2][3]
        self.vt_coeff.import_panel_value(self, (v[2][1],))

    def preprocess_params(self, engine):
        from petram.helper.init_helper import eval_value
        try:
            self.vt_coeff.preprocess_params(self)
            self.init_value = self.vt_coeff.make_value_or_expression(self)[0]
        except:
            self.init_value = 0.0
            assert False, traceback.format_exc()

    def get_phys(self):
        names = self.phys_model.split(',')
        names = [n.strip() for n in names if n.strip() != '']
        return [self.root()['Phys'][n] for n in names]

    def run(self, engine):
        phys_targets = self.get_phys()
        dwcparams = (self.name(),
                     self.init_dwc_name_params[0],
                     self.init_dwc_name_params[1])
        engine.run_apply_init0(phys_targets, self.init_mode,
                               init_value=self.init_value,
                               init_path=self.init_path,
                               init_var=self.init_var,
                               init_dwc=dwcparams)
        return phys_targets

    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('phys')


class CustomInitSetting():
    '''
    initialization w/o GUI
    usually called from Solver::get_custom_init()

    '''

    def __init__(self, phys_targets, var="", mode=1, value=[0, ], path='', dwc=('', ''), name='CustomInit'):
        self.init_name = name
        self.phys_targets = phys_targets
        self.init_mode = mode
        self.init_value = value
        self.init_path = path
        self.init_dwc_name_params = dwc
        self.init_var = var

    def name(self):
        return self.init_name

    def run(self, engine):
        dwcparams = (self.name,
                     self.init_dwc_name_params[0],
                     self.init_dwc_name_params[1])

        engine.run_apply_init0(self.phys_targets,
                               mode=self.init_mode,
                               init_value=self.init_value,
                               init_path=self.init_path,
                               init_var=self.init_var,
                               init_dwc=dwcparams)

        return self.phys_targets



import traceback
import numpy as np

from petram.model import Domain, Bdry, Edge, Point, Pair
from petram.phys.phys_model import VectorPhysCoefficient, PhysCoefficient
from petram.phys.weakform import WeakLinIntegration, WeakBilinIntegration

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('WF_Constraints')

from petram.mfem_config import use_parallel
if use_parallel:
   import mfem.par as mfem
else:
   import mfem.ser as mfem

# define variable for this BC.
from petram.phys.vtable import VtableElement, Vtable

'''
class InitValue(PhysCoefficient):
   def EvalValue(self, x):
       v = super(InitValue, self).EvalValue(x)
       if self.real:  val = v.real
       else: val =  v.imag
       return val
'''
'''    
   
class InitValueV(VectorPhysCoefficient):
   def EvalValue(self, x):
       v = super(InitValueV, self).EvalValue(x)
       if self.real:  val = v.real
       else: val =  v.imag
       return val
'''
class WF_common(object):
    can_timedpendent = True
    @property
    def vt3(self):
        names = self.get_root_phys().dep_vars_base
        names2 = [n + '_init' for n in names]
        data = []
        if hasattr(self, '_vt3'):
            vt3 = self._vt3
            if vt3.keys() == names2: return vt3
            
        data = [(n+'_init', VtableElement(n+'_init',
                                          type='array', guilabel = n+'(init)',
                                          default ="0.0", tip = "initial value",
                                          chkbox = True)) for n in names]
        self._vt3 = Vtable(data)
        v = {}
        self._vt3.attribute_set(v)
        self.do_update_attribute_set(v)
        return self._vt3
     
    def get_init_coeff(self, engine, real=True, kfes=0):
        names = self.get_root_phys().dep_vars_base
        
        if not getattr(self, 'use_'+names[kfes]+'_init'): return
        
        f_name = self.vt3.make_value_or_expression(self)

        el = self.get_root_phys().element
        if el.startswith('H1') or el.startswith('L2'):
            ll = self.get_root_phys().vdim
        else:
            ll = self.get_root_phys().ndim

        kwargs = {}

        from petram.phys.coefficient import SCoeff, VCoeff
        if ll == 1:
            coeff = SCoeff(f_name[0],
                       self.get_root_phys().ind_vars,
                       self._local_ns, self._global_ns,
                       real = real)
            
        else:
            coeff = VCoeff(ll, f_name[0],  self.get_root_phys().ind_vars,
                             self._local_ns, self._global_ns,
                             real = real)
            kwargs['vec'] = True
        return self.restrict_coeff(coeff, engine, **kwargs)

    
class WF_WeakDomainBilinConstraint(WF_common, Domain, WeakBilinIntegration):
    has_3rd_panel = True
    _has_4th_panel = True                
    def __init__(self, **kwargs):
        super(WF_WeakDomainBilinConstraint, self).__init__(**kwargs)
        Domain.__init__(self, **kwargs)
        WeakBilinIntegration.__init__(self)
        
    def attribute_set(self, v):
        Domain.attribute_set(self, v)
        WeakBilinIntegration.attribute_set(self, v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

    @classmethod
    def fancy_tree_name(self):
        return 'WeakContribution'

    @classmethod
    def fancy_menu_name(self):
        return 'WeakContribution'
     
class WF_WeakDomainLinConstraint(WF_common, Domain, WeakLinIntegration):
    has_3rd_panel = True
    def __init__(self, **kwargs):
        super(WF_WeakDomainLinConstraint, self).__init__(**kwargs)
        Domain.__init__(self, **kwargs)
        WeakLinIntegration.__init__(self)
        
    def attribute_set(self, v):
        Domain.attribute_set(self, v)
        WeakLinIntegration.attribute_set(self, v)         
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

    @classmethod
    def fancy_tree_name(self):
        return 'WeakLinearContribution'
     
    @classmethod
    def fancy_menu_name(self):
        return 'WeakLinearContribution'
        
class WF_WeakBdryBilinConstraint(WF_common, Bdry, WeakBilinIntegration):
    has_3rd_panel = True        
    def __init__(self, **kwargs):
        super(WF_WeakBdryBilinConstraint, self).__init__(**kwargs)        
        Bdry.__init__(self, **kwargs)
        WeakBilinIntegration.__init__(self)
        
    def attribute_set(self, v):
        Bdry.attribute_set(self, v)
        WeakBilinIntegration.attribute_set(self, v)                
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

    @classmethod
    def fancy_tree_name(self):
        return 'WeakContribution'

    @classmethod
    def fancy_menu_name(self):
        return 'WeakContribution'
     
class WF_WeakBdryLinConstraint(WF_common, Bdry, WeakLinIntegration):
    has_3rd_panel = True        
    def __init__(self, **kwargs):
        super(WF_WeakBdryLinConstraint, self).__init__(**kwargs)        
        Bdry.__init__(self, **kwargs)
        WeakLinIntegration.__init__(self)
        
    def attribute_set(self, v):
        Bdry.attribute_set(self, v)
        WeakLinIntegration.attribute_set(self, v)                
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v
     
    @classmethod
    def fancy_tree_name(self):
        return 'WeakLinearContribution'
     
    @classmethod
    def fancy_menu_name(self):
        return 'WeakLinearContribution'
     
class WF_WeakEdgeBilinConstraint(WF_common, Edge, WeakBilinIntegration):
    has_3rd_panel = True        
    def __init__(self, **kwargs):
        super(WF_WeakEdgeBilinConstraint, self).__init__(**kwargs)        
        Edge.__init__(self, **kwargs)
        WeakBilinIntegration.__init__(self)

    def attribute_set(self, v):
        Edge.attribute_set(self, v)
        WeakBilinIntegration.attribute_set(self, v)        
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

class WF_WeakEdgeLinConstraint(WF_common, Edge, WeakLinIntegration):
    has_3rd_panel = True        
    def __init__(self, **kwargs):
        super(WF_WeakEdgeLinConstraint, self).__init__(**kwargs)        
        Edge.__init__(self, **kwargs)
        WeakLinIntegration.__init__(self)

    def attribute_set(self, v):
        Edge.attribute_set(self, v)
        WeakLinIntegration.attribute_set(self, v)        
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v
     
from petram.phys.weakform import WeakLinDeltaIntegration, WeakBilinDeltaIntegration

class WF_WeakPointBilinConstraint(WF_common, Point, WeakBilinDeltaIntegration):
    has_3rd_panel = True        
    def __init__(self, **kwargs):
        super(WF_WeakPointBilinConstraint, self).__init__(**kwargs)        
        Point.__init__(self, **kwargs)
        WeakBilinDeltaIntegration.__init__(self)

    def attribute_set(self, v):
        Point.attribute_set(self, v)
        WeakBilinDeltaIntegration.attribute_set(self, v)
        v['sel_readonly'] = True
        v['sel_index'] = ['all']
        v['sel_index_txt'] = '["all"]'        
        return v
     
class WF_WeakPointLinConstraint(WF_common, Point, WeakLinDeltaIntegration):
    has_3rd_panel = True        
    def __init__(self, **kwargs):
        super(WF_WeakPointLinConstraint, self).__init__(**kwargs)        
        Point.__init__(self, **kwargs)
        WeakLinDeltaIntegration.__init__(self)

    def attribute_set(self, v):
        Point.attribute_set(self, v)
        WeakLinDeltaIntegration.attribute_set(self, v)
        v['sel_readonly'] = True
        v['sel_index'] = ['all']
        v['sel_index_txt'] = '["all"]'                        
        return v

     



'''
   BdrNodalEvaluator:
      a thing to evaluate solution on a boundary
'''
import numpy as np
import weakref
import six
import os

from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD

from petram.sol.evaluator_agent import EvaluatorAgent

class ProbeEvaluator(EvaluatorAgent):
    def __init__(self, battrs):
        super(ProbeEvaluator, self).__init__()
        self.battrs = battrs
        
    def preprocess_geometry(self,  *args, **kargs):
        pass

    def eval_probe(self, expr, xexpr, probe_files, phys):
        from petram.helper.variables import Variable, var_g
        from petram.sol.probe import load_probes

        #print("probe_files", probe_files)
        path = probe_files[0]
        path = os.path.expanduser(path)        
        probes = probe_files[1]

        code= compile(expr, '<string>', 'eval')
        names = list(code.co_names)

        if len(xexpr.strip()) != 0:
            xcode = compile(xexpr, '<string>', 'eval')
            names.extend(xcode.co_names)
        else:
            xcode = None

        g = phys._global_ns.copy()
        for key in var_g.keys():
            g[key] = var_g[key]

        default_xname = ''
        for n in names:
            if n in probes:
                xdata, ydata = load_probes(path, probes[n])
                g[n] = ydata
                for nn in xdata:
                    g[nn] = xdata[nn]
                    default_xname = nn
                    

        val = np.array(eval(code, g, {}), copy=False)
        if xcode is None:
            xval = g[default_xname]
        else:
            xval = np.array(eval(xcode, g, {}), copy=False)

        return xval, val




import numpy as np
import scipy
import six
import weakref
from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD


from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

from petram.sol.evaluator_agent import EvaluatorAgent
from petram.sol.bdr_nodal_evaluator import process_iverts2nodals
from petram.sol.bdr_nodal_evaluator import eval_at_nodals, get_emesh_idx

class SliceEvaluator(EvaluatorAgent):
    def __init__(self, attrs, plane = None):
        '''
           attrs = [1,2,3]
           plane = [ax, ay, az, c]

           cut-plane is defined as
           ax * x + ay * y + ax * z + c = 0
        '''
        super(SliceEvaluator, self).__init__()
        self.attrs = attrs
        self.plane = plane
        
    def preprocess_geometry(self, attrs, plane = None, emesh_idx=0):
        #from petram.sol.test import pg
        #return pg(self, battrs, plane = plane)
        self.vertices = None
        self.iverts = None

        self.knowns = WKD()
        mesh = self.mesh()[emesh_idx]
        self.iverts = []
        self.attrs = attrs
        self.plane = plane

        attrs = self.attrs
        axyz = self.plane[:3]
        c    = self.plane[-1]

        attr = mesh.GetAttributeArray()
        x = [np.where(attr == a)[0] for a in attrs]
        if np.sum([len(xx) for xx in x]) == 0: return

        ialleles = np.hstack(x).astype(int).flatten()
        ieles = []
        num_tri = 0
        tri_iverts = []
        f_values = []
        for iel in ialleles:
            verts = np.vstack([mesh.GetVertexArray(i)
                               for i in mesh.GetElement(iel).GetVerticesArray()])
            f = np.sum(verts*axyz, -1) + c

            ips = np.where(f > 0)[0]
            ims = np.where(f < 0)[0]
            izs = np.where(f == 0)[0]
            #print f, ips, ims, izs            
            if (len(ips) == 0 or len(ims) == 0) and len(izs) != 3:
                continue # outside the plane
            ieles.append(iel)
            f_values.append(f)

            if f.prod() > 0: # cross section is quad.
               num_tri += 2
            else:
               num_tri += 1

        # then get unique set of elements relating to the verts.
        if num_tri == 0:
            #print "not found"
            return
        print("found " + str(num_tri) + " elements")
        vert2el = mesh.GetVertexToElementTable()

        iverts = np.array([mesh.GetElement(iel).GetVerticesArray()
                              for iel in ieles])
        self.iverts = iverts

        data = process_iverts2nodals(mesh, iverts)
        for k in six.iterkeys(data):
            setattr(self, k, data[k])

        iverts_f = self.iverts_f

        # how to interpolate nodal values...
        vertices = np.zeros((num_tri, 3, 3))
        mat      = scipy.sparse.lil_matrix((num_tri*3, len(iverts_f)))

        def fill_vert_mat(vertics, mat, ivv, itri, k, i1, i2, verts, f):
            v1 = verts[i1]; v2 = verts[i2]
            f1 = np.abs(f[i1]); f2 = np.abs(f[i2])
            v = (v1*f2 + v2*f1)/(f1 + f2)
            vertices[itri, k, :] = v
            mat[itri*3+k, ivv[i1]] = np.abs(f2)/(f1+f2)
            mat[itri*3+k, ivv[i2]] = np.abs(f1)/(f1+f2)

        itri = 0
        for iel, f in zip(ieles, f_values):
            iv = mesh.GetElement(iel).GetVerticesArray()
            verts = np.vstack([mesh.GetVertexArray(i) for i in iv])
            ivv = [np.where(iverts_f == i)[0] for i in iv]
            ips = np.where(f > 0)[0]
            ims = np.where(f < 0)[0]           
            if (len(ips) == 2 and len(ims) == 2):
                fill_vert_mat(vertices, mat, ivv, itri, 0, ips[0], ims[0], verts, f)
                fill_vert_mat(vertices, mat, ivv, itri, 1, ips[0], ims[1], verts, f)
                fill_vert_mat(vertices, mat, ivv, itri, 2, ips[1], ims[0], verts, f)            
                itri += 1

                fill_vert_mat(vertices, mat, ivv, itri, 0, ips[1], ims[0], verts, f)
                fill_vert_mat(vertices, mat, ivv, itri, 1, ips[1], ims[1], verts, f)
                fill_vert_mat(vertices, mat, ivv, itri, 2, ips[0], ims[1], verts, f)            
                itri += 1
            else:

                if len(ips) == 1:
                    ims = np.where(f <= 0)[0]                                       
                    i =  ips[0]
                    ii = ims
                else:
                    ips = np.where(f >= 0)[0]   
                    i = ims[0]
                    ii = ips

                fill_vert_mat(vertices, mat, ivv, itri, 0, i, ii[0], verts, f)
                fill_vert_mat(vertices, mat, ivv, itri, 1, i, ii[1], verts, f)
                fill_vert_mat(vertices, mat, ivv, itri, 2, i, ii[2], verts, f)    
                itri += 1

        self.ibeles = None # can not use boundary variable in this evaulator        
        self.vertices = vertices
        self.interp_mat = mat

        self.emesh_idx = emesh_idx
        
    def set_plane(self, a, b, c, d):
        '''
        ax + by + cz + d = 0
        '''
        self.attrs[1] = a
        self.attrs[2] = b
        self.attrs[3] = c
        self.attrs[4] = d        

    def eval(self, expr, solvars, phys, **kwargs):
        emesh_idx = get_emesh_idx(self, expr, solvars, phys)
        if len(emesh_idx) > 1:
            assert False, "expression involves multiple mesh (emesh length != 1)"
        #if len(emesh_idx) < 1:
        #    assert False, "expression is not defined on any mesh"
        #(this could happen when expression is pure geometryical like "x+y")
            
        if len(emesh_idx) == 1:
            if self.emesh_idx != emesh_idx[0]:
                self.preprocess_geometry(self.attrs, plane=self.plane,
                                      emesh_idx=emesh_idx[0])
        
        if self.vertices is None: return None, None, None
        
        val = eval_at_nodals(self, expr, solvars, phys)
        if val is None: return None, None, None

        value = self.interp_mat.dot(val)
        value = value.flatten()
        verts = self.vertices.reshape(-1, self.vertices.shape[-1])
        return (verts, value, np.arange(len(value)).
                reshape(-1, self.vertices.shape[-1]))
        
    



from __future__ import print_function
from functools import reduce
'''
Utility class to handle BlockMatrix made from scipy-sparse and
Hypre with the same interface
'''
import numpy as np
import scipy
from scipy.sparse import coo_matrix, spmatrix, lil_matrix, csc_matrix

from petram.mfem_config import use_parallel
import mfem.common.chypre as chypre
from mfem.common.mpi_debug import nicePrint

if use_parallel:
    from petram.helper.mpi_recipes import *
    from mfem.common.parcsr_extra import *
    import mfem.par as mfem
    default_kind = 'hypre'
else:
    import mfem.ser as mfem
    default_kind = 'scipy'

from petram.solver.solver_utils import make_numpy_coo_matrix
from petram.helper.matrix_file import write_coo_matrix, write_vector

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('BlockMatrix')
format_memory_usage = debug.format_memory_usage


class One(object):
    '''
    An identity matrix (used in P and mimic 1*X = X)
    '''

    def __init__(self, ref):
        self._shape = ref.shape
        self._is_hypre = False
        if hasattr(ref, "GetColPartArray"):
            self._cpart = ref.GetColPartArray()
            self._is_hypre = True
        if hasattr(ref, "GetRowPartArray"):
            self._rpart = ref.GetRowPartArray()
            self._is_hypre = True

    def __repr__(self):
        return "IdentityMatrix"

    @property
    def shape(self): return self._shape

    @property
    def nnz(self):
        return self.shape[0]

    def true_nnz(self):
        return self.shape[0]

    def GetColPartArray(self):
        return self._cpart

    def GetRowPartArray(self):
        return self._rpart

    def __mul__(self, other):
        return other

    def transpose(self):
        return self

    def dot(self, other):
        return other

    @property
    def isHypre(self):
        return self._is_hypre


class ScipyCoo(coo_matrix):
    def true_nnz(self):
        if hasattr(self, "eliminate_zeros"):
            self.eliminate_zeros()
        return self.nnz

    def __add__(self, other):
        ret = super(ScipyCoo, self).__add__(other)
        return convert_to_ScipyCoo(ret)

    def __sub__(self, other):
        ret = super(ScipyCoo, self).__sub__(other)
        return convert_to_ScipyCoo(ret)

    def setDiag(self, idx, value=1.0):

        ret = self.tolil()
        idx = np.array(idx, dtype=int, copy=False)
        ret[idx, idx] = value
        # for i in idx:
        #   ret[i,i] = value
        ret = ret.tocoo()
        self.data = ret.data
        self.row = ret.row
        self.col = ret.col
    '''
    def resetDiagImag(self, idx):
        ret = self.tolil()

        for i in idx:
           ret[i,i] = ret[i,i].real
        ret = ret.tocoo()
        self.data = ret.data
        self.row  = ret.row
        self.col  = ret.col
    '''

    def resetRow(self, rows, inplace=True):
        rows = np.array(rows, dtype=int, copy=False)
        #ret = self.tolil()
        #ret[rows, :] = 0.0
        #ret.data[np.in1d(ret.row in rows)] = 0.0
        ret = self.tocsr()
        ret[rows, :] = ret[rows, :] * 0.0
        ret = ret.tocoo()

        if inplace:
            self.data = ret.data
            self.row = ret.row
            self.col = ret.col
            return self
        else:
            return ret

    def resetCol(self, cols, inplace=True):
        cols = np.array(cols, dtype=int, copy=False)
        #ret.data[np.in1d(ret.col in cols)] = 0.0
        #ret = self.tolil()
        ret = self.tocsc()
        ret[:, cols] = ret[:, cols] * 0.0
        ret = ret.tocoo()

        if inplace:
            self.data = ret.data
            self.row = ret.row
            self.col = ret.col
            return self
        else:
            return ret

    def selectRows(self, nonzeros):
        m = self.tocsr()
        ret = (m[nonzeros, :]).tocoo()
        return convert_to_ScipyCoo(ret)

    def selectCols(self, nonzeros):
        m = self.tocsc()
        ret = (m[:, nonzeros]).tocoo()
        return convert_to_ScipyCoo(ret)

    def dot(self, P):
        return convert_to_ScipyCoo(super(ScipyCoo, self).dot(P))

    def rap(self, P):
        PP = P.conj().transpose()
        return convert_to_ScipyCoo(PP.dot(self.dot(P)))

    def conj(self, inplace=False):
        if inplace:
            np.conj(self.data, out=self.data)
            return self
        else:
            return self.conj()

    def elimination_matrix(self, nonzeros):
        '''
        # P elimination matrix for column vector
        [1 0  0][x1]    [x1]
        [0 0  1][x2]  = [x3]
                [x3]

        P^t (transpose does reverse operation)
           [1 0][x1]    [x1]
           [0 0][x3]  = [0]
           [0,1]        [x3]

        # P^t does elimination matrix for horizontal vector
                  [1,0]
        [x1 x2 x3][0,0]  = [x1, x3]
                  [0,1]

        # P does reverse operation for horizontal vector
               [1,0 0]
        [x1 x3][0,0 1]  = [x1, 0  x3]
        '''
        ret = lil_matrix((len(nonzeros), self.shape[0]))
        for k, z in enumerate(nonzeros):
            ret[k, z] = 1.
        return convert_to_ScipyCoo(ret.tocoo())

    def get_global_coo(self):
        '''
        global representation:
           zero on non-root node
        '''
        try:
            from mpi4py import MPI
            myid = MPI.COMM_WORLD.rank
            if myid != 0:
                return coo_matrix(self.shape)
            else:
                return self
        except BaseException:
            return self

    def get_mfem_sparsemat(self):
        '''
        generate mfem::SparseMatrix using the same data
        '''
        if np.iscomplexobj(self):
            csr_r = self.tocsr().real
            csr_r.eliminate_zeros()
            csr_i = self.tocsr().imag
            csr_i.eliminate_zeros()
            return mfem.SparseMatrix(csr_r), mfem.SparseMatrix(csr_i)
        else:
            csr_r = self.tocsr()
            csr_r.eliminate_zeros()
            csr_i = None
            return mfem.SparseMatrix(csr_r), None

    def __repr__(self):
        return "ScipyCoo" + str(self.shape)

    def __str__(self):
        return "ScipyCoo" + str(self.shape) + super(ScipyCoo, self).__str__()

    @property
    def isHypre(self):
        return False

    '''
    two function to provde the same interface as CHypre
    '''

    def GetColPartArray(self):
        return (0, self.shape[1], self.shape[1])

    def GetRowPartArray(self):
        return (0, self.shape[0], self.shape[0])
    GetPartitioningArray = GetRowPartArray

    def eliminate_RowsCols(self, B, tdof, inplace=True, diagpolicy=1):
        '''
        daigpolicy = 0  # DiagOne
        daigpolicy = 1  # DiagKeep

        Note: policy is controled from engine::filL_BCeliminate_matrix
        '''
        print("inplace flag off copying matrix")
        # A + Ae style elimination
        idx = np.in1d(self.col, tdof)
        idx2 = np.in1d(self.row, tdof)

        if diagpolicy == 0:
            diagAe = self.diagonal()[tdof] - 1
            diagA = 1
        else:
            diagAe = 0
            diagA = self.diagonal()[tdof]

        aidx = np.logical_or(idx, idx2)
        #aidx = []
        AeCol = self.col[aidx]
        AeRow = self.row[aidx]
        AeData = self.data[aidx]
        Ae2 = coo_matrix((AeData, (AeRow, AeCol)),
                         shape=self.shape, dtype=self.dtype)
        Ae2c = Ae2.tocsr()
        Ae2c[tdof, tdof] = diagAe
        Ae2 = Ae2c.tocoo()

        if inplace:
            target = self
            target_b = B.tolil()
        else:
            target = self.copy()
            target_b = B.copy().tolil()

        target.data[idx] = 0
        target.data[idx2] = 0
        target.eliminate_zeros()
        lil2 = target.tolil()
        lil2[tdof, tdof] = diagA

        # if diagpolicy == 1:
        # target_b[tdof, 0] = target_b[tdof, 0].toarray().flatten() * diagA
        target_b[tdof, 0] = diagA

        coo = lil2.tocoo()
        target.data = coo.data
        target.row = coo.row
        target.col = coo.col

        coo_b = convert_to_ScipyCoo(target_b)

        return Ae2, target, coo_b

    def get_elements(self, tdof):
        slil = self.tolil()
        value = slil[tdof, :].toarray()
        return value

    def set_elements(self, tdof, m):
        slil = self.tolil()
        slil[tdof, :1] = m
        coo = slil.tocoo()
        self.data = coo.data
        self.row = coo.row
        self.col = coo.col

    def copy_element(self, tdof, m):
        mlil = m.tolil()
        value = mlil[tdof, 0].toarray()

        slil = self.tolil()
        slil[tdof, :1] = value
        coo = slil.tocoo()
        self.data = coo.data
        self.row = coo.row
        self.col = coo.col


def convert_to_ScipyCoo(mat):
    if isinstance(mat, np.ndarray):
        mat = coo_matrix(mat)
    if isinstance(mat, spmatrix):
        if not isinstance(mat, coo_matrix):
            mat = mat.tocoo()
        mat.__class__ = ScipyCoo
    return mat


class BlockMatrix(object):
    def __init__(self, shape, kind=default_kind, complex=False):
        '''
        kind : scipy
                  stores scipy sparse or numpy array
               hypre
        '''
        self.block = [[None] * shape[1] for x in range(shape[0])]
        self.kind = kind
        self.shape = shape
        self.complex = complex

    def __getitem__(self, idx):
        try:
            r, c = idx
        except BaseException:
            r = idx
            c = 0
        if isinstance(r, slice):
            new_r = range(self.shape[0])[r]
            ret = BlockMatrix((len(new_r), self.shape[1]))
            for i, r in enumerate(new_r):
                for j in range(self.shape[1]):
                    ret[i, j] = self[r, j]
            return ret
        elif isinstance(c, slice):
            new_c = range(self.shape[1])[c]
            ret = BlockMatrix((self.shape[0], len(new_c)))
            for i in range(self.shape[0]):
                for j, c in enumerate(new_c):
                    ret[i, j] = self[i, c]
            return ret
        else:
            return self.block[r][c]

    def __setitem__(self, idx, v):
        try:
            r, c = idx
        except BaseException:
            r = idx
            c = 0
        if v is not None:
            if isinstance(v, chypre.CHypreMat):
                if v.isComplex():
                    self.complex = True
            elif isinstance(v, chypre.CHypreVec):
                if v.isComplex():
                    self.complex = True
            elif v is None:
                pass
            else:
                v = convert_to_ScipyCoo(v)
                if np.iscomplexobj(v):
                    self.complex = True
        self.block[r][c] = v

    def __add__(self, v):
        if self.shape != v.shape:
            raise ValueError("Block format is inconsistent")

        shape = self.shape
        complex = self.complex or v.complex
        ret = BlockMatrix(shape, kind=self.kind, complex=complex)
        for i in range(shape[0]):
            for j in range(shape[1]):
                if self[i, j] is None and v[i, j] is None:
                    ret[i, j] = None
                elif self[i, j] is None:
                    ret[i, j] = v[i, j]
                elif v[i, j] is None:
                    ret[i, j] = self[i, j]
                else:
                    ret[i, j] = self[i, j] + v[i, j]
        return ret

    def __sub__(self, v):
        if self.shape != v.shape:
            raise ValueError("Block format is inconsistent")
        shape = self.shape
        complex = self.complex or v.complex
        ret = BlockMatrix(shape, kind=self.kind, complex=complex)

        for i in range(shape[0]):
            for j in range(shape[1]):
                if self[i, j] is None and v[i, j] is None:
                    ret[i, j] = None
                elif self[i, j] is None:
                    ret[i, j] = -v[i, j]
                elif v[i, j] is None:
                    ret[i, j] = self[i, j]
                else:
                    ret[i, j] = self[i, j] - v[i, j]
        return ret

    def __mul__(self, other):  # note other is scalar.. this is not matrix multiplicaiton
        shape = self.shape

        ret = BlockMatrix(shape, kind=self.kind, complex=self.complex)

        for i in range(shape[0]):
            for j in range(shape[1]):
                if self[i, j] is not None:
                    ret[i, j] = self[i, j] * other
        return ret

    def __neg__(self):
        shape = self.shape
        ret = BlockMatrix(shape, kind=self.kind, complex=self.complex)

        for i in range(shape[0]):
            for j in range(shape[1]):
                if self[i, j] is not None:
                    ret[i, j] = -self[i, j]
        return ret

    def __repr__(self):
        txt = ["BlockMatrix" + str(self.shape)]
        for i in range(self.shape[0]):
            txt.append(str(i) + " : " + "  ".join([self.block[i][j].__repr__()
                                                   for j in range(self.shape[1])]))
        return "\n".join(txt) + "\n"

    def format_nnz(self):
        txt = []
        for i in range(self.shape[0]):
            txt.append(str(i) + " : " + ",  ".join([str(self[i, j].nnz)
                                                    if hasattr(self[i, j], "nnz") else "unknown"
                                                    for j in range(self.shape[1])]))
        return "\nnon-zero elements (nnz)\n" + "\n".join(txt)

    def normsq(self):
        shape = self.shape
        assert shape[1] == 1, "multilpe vectors are not supported"
        norm = [0]*shape[0]

        for i in range(shape[0]):
            # for j in range(shape[1]):
            for j in [0]:
                v = self[i, j]
                if isinstance(v, chypre.CHypreVec):
                    vec = v.toarray()
                elif isinstance(v, ScipyCoo):
                    vec = v.toarray()
                else:
                    assert False, "not supported"
                norm[i] += np.abs(np.sum(vec*np.conj(vec)))

        if use_parallel:
            norm = np.sum(allgather(norm), 0)

        return norm

    def norm(self):
        return np.sqrt(self.normsq())

    def average_norm(self, sq=False):
        shape = self.shape
        assert shape[1] == 1, "multilpe vectors are not supported"

        if sq:
            norm = self.normsq()
        else:
            norm = self.norm()

        length = [0]*shape[0]

        for i in range(shape[0]):
            for j in range(shape[1]):
                v = self[i, j]
                length[i] += np.prod(v.shape)  # shape is (1,x) or (x, 1)

        if use_parallel:
            length = np.sum(allgather(length), 0)

        length = np.array(length)
        return norm/length

    def print_nnz(self):
        print(self.format_nnz())

    def format_true_nnz(self):
        txt = []
        for i in range(self.shape[0]):
            txt.append(str(i) + " : " + ",  ".join([str(self[i, j].true_nnz())
                                                    if hasattr(self[i, j], "true_nnz") else "unknown"
                                                    for j in range(self.shape[1])]))
        return "non-zero elements (true nnz)\n" + "\n".join(txt)

    def print_true_nnz(self):
        print(self.format_ture_nnz())

    def print_row_part(self):
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                if self[i, j] is not None:
                    print(i, j, self[i, j].GetRowPartArray())

    def print_col_part(self):
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                if self[i, j] is not None:
                    print(i, j, self[i, j].GetColPartArray())

    def get_local_row_height(self, i):
        for j in range(self.shape[1]):
            blk = self.block[i][j]
            if blk is not None:
                return blk.shape[0]
        return None

    def get_local_col_width(self, j):
        for i in range(self.shape[0]):
            blk = self.block[i][j]
            if blk is not None:
                return blk.shape[1]
        return None

    def get_local_row_heights(self):
        return [self.get_local_row_height(i) for i in range(self.shape[0])]

    def get_local_col_widths(self):
        return [self.get_local_col_width(j) for j in range(self.shape[1])]

    def save_to_file(self, file):
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                name = file + '_' + str(i) + '_' + str(j)
                v = self[i, j]
                if isinstance(v, chypre.CHypreMat):
                    m = v.get_local_coo()
                    write_coo_matrix(name, m)
                elif isinstance(v, chypre.CHypreVec):
                    m = coo_matrix(v.toarray()).transpose()
                    write_coo_matrix(name, m)
                elif isinstance(v, ScipyCoo):
                    write_coo_matrix(name, v)
                elif v is None:
                    continue
                else:
                    assert False, "Don't know how to write file for " + \
                        str(type(v))

    def transpose(self):
        ret = BlockMatrix((self.shape[1], self.shape[0]), kind=self.kind)
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                if self[i, j] is not None:
                    ret[j, i] = self[i, j].transpose()
        return ret

    def add_to_element(self, i, j, v):
        if self[i, j] is None:
            self[i, j] = v
        else:
            self[i, j] = self[i, j] + v

    def dot(self, mat):
        if self.shape[1] != mat.shape[0]:
            raise ValueError("Block format is inconsistent")

        shape = (self.shape[0], mat.shape[1])
        ret = BlockMatrix(shape, kind=self.kind)

        for i in range(shape[0]):
            for j in range(shape[1]):
                for k in range(self.shape[1]):
                    if self[i, k] is None:
                        continue
                    elif mat[k, j] is None:
                        continue
                    elif ret[i, j] is None:
                        ret[i, j] = self[i, k].dot(mat[k, j])
                    else:
                        # print "dot is called here", self[i, k], mat[k, j]
                        ret[i, j] = ret[i, j] + self[i, k].dot(mat[k, j])
                    # try:
                    #    ret[i,j].shape
                    # except:
                    #    ret[i,j] = coo_matrix([[ret[i,j]]])
        return ret

    def get_subblock(self, mask1, mask2):
        ret = BlockMatrix((sum(mask1), sum(mask2)), kind=self.kind,
                          complex=self.complex)
        imask1 = [x for x in range(len(mask1)) if mask1[x]]
        imask2 = [x for x in range(len(mask2)) if mask2[x]]
        for i, ii in enumerate(imask1):
            for j, jj in enumerate(imask2):
                ret[i, j] = self[ii, jj]
        return ret

    def get_block_id(self, ignore_none=True):
        blocks = sum(self.block, [])
        if ignore_none:
            return [id(b) for b in blocks if b is not None]
        else:
            return [id(b) for b in blocks]

    def check_shared_id(self, target):
        ids = target.get_block_id()
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                if self[i, j] is None:
                    continue
                if id(self[i, j]) in ids:
                    print(i, j, "shared")

    def eliminate_empty_rowcol(self):
        '''
        collect empty row first. (no global communicaiton this step)

        share empty rows..

        apply it to all node

        '''
        from functools import reduce
        ret = BlockMatrix(self.shape, kind=self.kind)
        P2 = BlockMatrix(self.shape, kind=self.kind)

        dprint1(self.format_true_nnz())

        for i in range(self.shape[0]):
            nonzeros = []
            mat = None
            for j in range(self.shape[1]):
                if self[i, j] is not None:
                    if isinstance(self[i, j], ScipyCoo):
                        coo = self[i, j]
                        csr = coo.tocsr()
                        num_nonzeros = np.diff(csr.indptr)
                        knonzero = np.where(num_nonzeros != 0)[0]
                        if mat is None:
                            mat = coo
                    elif isinstance(self[i, j], chypre.CHypreMat):
                        coo = self[i, j].get_local_coo()
                        if hasattr(coo, "eliminate_zeros"):
                            coo.eliminate_zeros()
                        csr = coo.tocsr()
                        num_nonzeros = np.diff(csr.indptr)
                        knonzero = np.where(num_nonzeros != 0)[0]
                        if mat is None:
                            mat = self[i, j]
                    elif isinstance(self[i, j], chypre.CHypreVec):
                        if self[i, j].isAllZero():
                            knonzero = []
                        else:
                            knonzero = [0]
                    elif (isinstance(self[i, j], np.ndarray) and
                          self[i, j].ndim == 2):
                        knonzero = [k for k in range(self[i, j].shape[0])
                                    if any(self[i, j][k, :])]
                        self[i, j] = convert_to_ScipyCoo(self[i, j])
                        mat = self[i, j]
                    else:
                        raise ValueError("Unsuported Block Element" +
                                         str(type(self[i, j])))
                    nonzeros.append(knonzero)
                else:
                    nonzeros.append([])
            knonzeros = reduce(np.union1d, nonzeros)
            knonzeros = np.array(knonzeros, dtype=np.int64)
            # share nonzero to eliminate column...
            if self.kind == 'hypre':
                if isinstance(self[i, i], chypre.CHypreMat):
                    gknonzeros = self[i, i].GetRowPartArray()[0] + knonzeros
                else:
                    gknonzeros = knonzeros
                gknonzeros = allgather_vector(gknonzeros)
                gknonzeros = np.unique(gknonzeros)
                dprint2('nnz', coo.nnz, len(knonzero))
            else:
                gknonzeros = knonzeros

            if mat is not None:
                # hight and row partitioning of mat is used
                # to construct P2
                if len(gknonzeros) < self[i, i].shape[0]:
                    P2[i, i] = self[i, i].elimination_matrix(gknonzeros)
                else:
                    P2[i, i] = One(self[i, i])
            # what if common zero rows differs from common zero col?
            for j in range(self.shape[1]):
                if ret[i, j] is not None:
                    ret[i, j] = ret[i, j].selectRows(gknonzeros)
                elif self[i, j] is not None:
                    ret[i, j] = self[i, j].selectRows(gknonzeros)
                else:
                    pass
                if ret[j, i] is not None:
                    ret[j, i] = ret[j, i].selectCols(gknonzeros)
                elif self[j, i] is not None:
                    ret[j, i] = self[j, i].selectCols(gknonzeros)
                else:
                    pass

        return ret, P2

    def reformat_distributed_mat(self, mat, ksol, ret, mask, alpha=1, beta=0):
        '''
        reformat distributed matrix into blockmatrix (columne vector)
        '''
        L = []
        idx = 0
        imask = [x for x in range(len(mask[0])) if mask[0][x]]
        jmask = [x for x in range(len(mask[1])) if mask[1][x]]

        for j in jmask:
            for i in imask:
                if self[i, j] is not None:
                    if self.kind == 'scipy':
                        l = self[i, j].shape[1]
                    else:
                        part = self[i, j].GetColPartArray()
                        l = part[1] - part[0]
                    ref = self[i, j]
                    break
            L.append(l)
            v = mat[idx:idx + l, ksol]
            idx = idx + l

            ret.set_element_from_distributed_mat(
                v, j, 0, ref, alpha=alpha, beta=beta)
        return ret

    def set_element_from_distributed_mat(self, v, i, j, ref, alpha=1, beta=0):
        if self.kind == 'scipy':
            if alpha == 1 and beta == 0:
                self[i, j] = v.reshape(-1, 1)
            else:
                self[i, j] = self[i, j]*beta + alpha * v.reshape(-1, 1)
        else:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD

            if ref.isHypre:
                v = np.ascontiguousarray(v)
                if np.iscomplexobj(v):
                    if alpha == 1 and beta == 0:
                        rv = ToHypreParVec(v.real)
                        iv = ToHypreParVec(v.imag)
                        self[i, j] = chypre.CHypreVec(rv, iv)
                    else:
                        self[i, j] *= beta
                        realpart = self[i, j][0].GetDataArray()
                        realpart += v.real*alpha
                        imagpart = self[i, j][1].GetDataArray()
                        imagpart += v.imag*alpha
                else:
                    rv = ToHypreParVec(v)
                    if alpha == 1 and beta == 0:
                        self[i, j] = chypre.CHypreVec(rv, None)
                    else:
                        self[i, j] *= beta
                        realpart = self[i, j][0].GetDataArray()
                        realpart += v*alpha
            else:
                assert False, "bug. this mode is not supported"

    def reformat_central_mat(self, mat, ksol, ret, mask, alpha=1, beta=0):
        '''
        reformat central matrix into blockmatrix (columne vector)
        so that matrix can be multiplied from the right of this

        self is a block diagonal matrix

        by default, ret is replaced by a new data
        with alpha != 1 or beta!=0, it will be set to 
             old data * beta + new_data * alpha
        '''
        L = []
        idx = 0
        imask = [x for x in range(len(mask[0])) if mask[0][x]]
        jmask = [x for x in range(len(mask[1])) if mask[1][x]]

        for j in jmask:
            for i in imask:
                if self[i, j] is not None:
                    l = self[i, j].shape[1]
                    break
            L.append(l)
            ref = self[i, j]

            if mat is not None:
                v = mat[idx:idx + l, ksol]
            else:
                v = None   # slave node (will recive data)
            idx = idx + l
            ret.set_element_from_central_mat(
                v, j, 0, ref, alpha=alpha, beta=beta)
        return ret

    def set_element_from_central_mat(self, v, i, j, ref, alpha=1, beta=0):
        '''
        set element using vector in root node
        row partitioning is taken from column partitioning
        of ref
        '''
        if self.kind == 'scipy':
            #print("here", type(self[i,j]))
            # if isinstance(self[i,j], ScipyCoo):
            #   print("here", self[i,j].shape, self[i,j])
            if alpha == 1 and beta == 0:
                self[i, j] = v.reshape(-1, 1)
            else:
                self[i, j] = self[i, j]*beta + alpha * v.reshape(-1, 1)

        else:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD

            if ref.isHypre:
                from mpi4py import MPI
                comm = MPI.COMM_WORLD

                part = ref.GetColPartArray()
                v = comm.bcast(v)
                start_col = part[0]
                end_col = part[1]

                v = np.ascontiguousarray(v[start_col:end_col])
                if np.iscomplexobj(v):
                    if alpha == 1 and beta == 0:
                        rv = ToHypreParVec(v.real)
                        iv = ToHypreParVec(v.imag)
                        self[i, j] = chypre.CHypreVec(rv, iv)
                    else:
                        self[i, j] *= beta
                        realpart = self[i, j][0].GetDataArray()
                        realpart += v.real*alpha
                        imagpart = self[i, j][1].GetDataArray()
                        imagpart += v.imag*alpha
                else:
                    if alpha == 1 and beta == 0:
                        rv = ToHypreParVec(v)
                        self[i, j] = chypre.CHypreVec(rv, None)
                    else:
                        self[i, j] *= beta
                        realpart = self[i, j][0].GetDataArray()
                        realpart += v*alpha
            else:
                # slave node gets the copy
                v = comm.bcast(v)
                if alpha == 1 and beta == 0:
                    self[i, j] = v.reshape(-1, 1)
                else:
                    self[i, j] = self[i, j]*beta + alpha * v.reshape(-1, 1)

    def get_squaremat_from_right(self, r, c):
        size = self[r, c].shape
        if self.kind == 'scipy':
            return ScipyCoo((size[1], size[1]))
        else:
            # this will return CHypreMat
            return self[r, c].get_squaremat_from_right()
    #
    #  methods for coo format
    #

    def gather_densevec(self):
        '''
        gather vector data to head node as dense data (for rhs)
        '''
        if self.kind == 'scipy':
            if self.complex:
                M = scipy.sparse.bmat(self.block, format='coo',
                                      dtype='complex').toarray()
            else:
                M = scipy.sparse.bmat(self.block, format='coo',
                                      dtype='float').toarray()
            return M
        else:
            data = []
            for i in range(self.shape[0]):
                if isinstance(self[i, 0], chypre.CHypreVec):
                    data.append(self[i, 0].GlobalVector())
                elif isinstance(self[i, 0], np.ndarray):
                    data.append(self[i, 0].flatten())
                elif isinstance(self[i, 0], ScipyCoo):
                    data.append(self[i, 0].toarray().flatten())
                else:
                    raise ValueError("Unsupported element" + str((i, 0)) +
                                     ":" + str(type(self[i, 0])))
            return np.hstack(data).reshape(-1, 1)

    def add_empty_square_block(self, r, c):
        '''
        add a square matrix to r, c element

        note: this is used to add diagnal block so that essential BC elimination
              works when diagnaol block is not specified in the model tree

              it is assuemd that the size of matrix can be inferred from
              off-diagonal blocks

              It generate a matrix whose memory for diag elements is allocated.

              If the size cannot be inferred, it does not do anything. This
              means form has zero col/row blocks. This happens when time-dependent
              solver fills the matrix (row corresponding to d/dt)
        '''
        #dprint1("adding empty block ", r, c)
        if self[r, c] is not None:
            assert False, "block is already filled"
        roffset = []
        coffset = []
        rsize = -1
        csize = -1
        for i in range(self.shape[0]):
            if self[i, c] is not None:
                csize = self[i, c].shape[1]
                if self.kind != 'scipy':
                    cp = self[i, c].GetColPartArray()
                    shape = self[i, c].shape
        for j in range(self.shape[1]):
            if self[r, j] is not None:
                rsize = self[r, j].shape[0]
                if self.kind != 'scipy':
                    rp = self[r, j].GetRowPartArray()
                    gsize = self[r, j].shape[0]

        if rsize < 0 or csize < 0:
            # this case does not fill the dialgona matrix
            return

        if rsize != csize:
            assert False, "matrix is not squre"
        if self.kind == 'scipy':
            m = scipy.sparse.eye(rsize, format='coo', dtype=float) * 0
            self[r, c] = convert_to_ScipyCoo(m)
        else:
            from mfem.common.chypre import SquareCHypreMat
            m = SquareCHypreMat(gsize, rp, real=True)
            self[r, c] = m

    def get_global_offsets(self, convert_real=False,
                           interleave=True):
        '''
        build matrix in coordinate format
        '''
        roffset = []
        coffset = []
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                if self[i, j] is not None:
                    roffset.append(self[i, j].shape[0])
                    break
        for j in range(self.shape[1]):
            for i in range(self.shape[0]):
                if self[i, j] is not None:
                    coffset.append(self[i, j].shape[1])
                    break
        #coffset = [self[0, j].shape[1] for j in range(self.shape[1])]
        if self.complex and convert_real:
            if interleave:
                roffset = np.vstack((roffset, roffset)).flatten()
                coffset = np.vstack((roffset, roffset)).flatten()
            else:
                roffset = np.hstack((roffset, roffset))
                coffset = np.hstack((roffset, roffset))

        roffsets = np.hstack([0, np.cumsum(roffset)])
        coffsets = np.hstack([0, np.cumsum(coffset)])
        return roffsets, coffsets

    def get_global_coo(self, dtype='float'):
        roffsets, coffsets = self.get_global_offsets()
        col = []
        row = []
        data = []
        glcoo = coo_matrix((roffsets[-1], coffsets[-1]), dtype=dtype)
        dprint1("roffset(get_global_coo)", roffsets)
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                if self[i, j] is None:
                    continue
                gcoo = self[i, j].get_global_coo()
                row.append(gcoo.row + roffsets[i])
                col.append(gcoo.col + coffsets[j])
                data.append(gcoo.data)
        glcoo.col = np.hstack(col)
        glcoo.row = np.hstack(row)
        glcoo.data = np.hstack(data)

        return glcoo

    #
    #  methods for distributed csr format
    #
    def get_local_partitioning(self, convert_real=True,
                               interleave=True,
                               merge_realimag=False):
        '''
        build matrix in coordinate format
        '''
        roffset = np.zeros(self.shape[0], dtype=int)
        coffset = np.zeros(self.shape[1], dtype=int)
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                if self[i, j] is not None:
                    rp = self[i, j].GetRowPartArray()
                    if (roffset[i] != 0 and
                            roffset[i] != rp[1] - rp[0]):
                        assert False, 'row partitioning is not consistent'
                    roffset[i] = rp[1] - rp[0]
                    if use_parallel and not isinstance(
                            self[i, j], chypre.CHypreMat):
                        from mpi4py import MPI
                        myid = MPI.COMM_WORLD.rank
                        if myid != 0:
                            roffset[i] = 0
        for j in range(self.shape[1]):
            for i in range(self.shape[0]):
                if self[i, j] is not None:
                    cp = self[i, j].GetColPartArray()
                    if (coffset[j] != 0 and
                            coffset[j] != cp[1] - cp[0]):
                        assert False, 'col partitioning is not consistent'
                    coffset[j] = cp[1] - cp[0]
                    if use_parallel and not isinstance(
                            self[i, j], chypre.CHypreMat):
                        if myid != 0:
                            coffset[i] = 0

        #coffset = [self[0, j].shape[1] for j in range(self.shape[1])]
        if self.complex and convert_real:
            if merge_realimag:
                roffset = roffset * 2
                coffset = coffset * 2
            else:
                if interleave:
                    roffset = np.repeat(roffset, 2)
                    coffset = np.repeat(coffset, 2)
                else:
                    roffset = np.hstack((roffset, roffset))
                    coffset = np.hstack((coffset, coffset))

        roffsets = np.hstack([0, np.cumsum(roffset)])
        coffsets = np.hstack([0, np.cumsum(coffset)])
        return roffsets, coffsets

    def get_local_partitioning_v(self, convert_real=True,
                                 interleave=True, size_hint=None):
        '''
        build matrix in coordinate format
        '''
        roffset = np.zeros(self.shape[0], dtype=int)
        if size_hint is not None:
            hint_shape = size_hint.shape
            # print "called with hint:", hint_shape, self.shape
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                if self[i, j] is not None:
                    rp = self[i, j].GetPartitioningArray()
                    if (roffset[i] != 0 and
                            roffset[i] != rp[1] - rp[0]):
                        assert False, 'row partitioning is not consistent'
                    roffset[i] = rp[1] - rp[0]
                elif size_hint is not None:
                    for k in range(hint_shape[1]):
                        if size_hint[i, k] is not None:
                            rp = size_hint[i, k].GetPartitioningArray()
                            if (roffset[i] != 0 and
                                    roffset[i] != rp[1] - rp[0]):
                                assert False, 'row partitioning is not consistent'
                            roffset[i] = rp[1] - rp[0]

        #coffset = [self[0, j].shape[1] for j in range(self.shape[1])]
        if self.complex and convert_real:
            if interleave:
                roffset = np.repeat(roffset, 2)
            else:
                roffset = np.hstack((roffset, roffset))

        roffsets = np.hstack([0, np.cumsum(roffset)])
        return roffsets

    def gather_blkvec_interleave(self, size_hint=None):
        '''
        Construct MFEM::BlockVector

        This routine ordered unkonws in the following order
           Re FFE1, Im FES1, ReFES2, Im FES2, ...
        If self.complex is False, it assembles a nomal block
           vector
        This routine is used together with get_global_blkmat_interleave(self):
        '''

        roffsets = self.get_local_partitioning_v(convert_real=True,
                                                 interleave=True,
                                                 size_hint=size_hint)
        dprint1("roffsets(vector)", roffsets)
        offset = mfem.intArray(list(roffsets))

        vec = mfem.BlockVector(offset)
        vec._offsets = offset  # in order to keep it from freed

        data = []
        ii = 0
        jj = 0

        # Here I don't like that I am copying the data between two vectors..
        # But, avoiding this takes the large rearangement of program flow...
        for i in range(self.shape[0]):
            if self[i, 0] is not None:
                if isinstance(self[i, 0], chypre.CHypreVec):
                    vec.GetBlock(ii).Assign(self[i, 0][0].GetDataArray())
                    if self.complex:
                        if self[i, 0][1] is not None:
                            vec.GetBlock(ii +
                                         1).Assign(self[i, 0][1].GetDataArray())
                        else:
                            vec.GetBlock(ii + 1).Assign(0.0)
                elif isinstance(self[i, 0], ScipyCoo):
                    arr = np.atleast_1d(self[i, 0].toarray().squeeze())
                    vec.GetBlock(ii).Assign(np.real(arr))
                    if self.complex:
                        vec.GetBlock(ii + 1).Assign(np.imag(arr))
                else:
                    assert False, "not implemented, " + str(type(self[i, 0]))
            else:
                vec.GetBlock(ii).Assign(0.0)
                if self.complex:
                    vec.GetBlock(ii + 1).Assign(0.0)
            ii = ii + 2 if self.complex else ii + 1

        return vec

    def get_global_blkmat_interleave(self):
        '''
        This routine ordered unkonws in the following order
           FFE1, FES2,
        If it is complex
           Re FFE1, Im FES1, ReFES2, Im FES2, ...
        '''
        roffsets, coffsets = self.get_local_partitioning(convert_real=True,
                                                         interleave=True)
        dprint1("Generating MFEM BlockMatrix: shape = " +
                str((len(roffsets) - 1, len(coffsets) - 1)))
        dprint1(
            "Generating MFEM BlockMatrix: roffset/coffset = ",
            roffsets,
            coffsets)

        ro = mfem.intArray(list(roffsets))
        co = mfem.intArray(list(coffsets))
        glcsr = mfem.BlockOperator(ro, co)

        ii = 0

        for i in range(self.shape[0]):
            jj = 0
            for j in range(self.shape[1]):
                if self[i, j] is not None:
                    if use_parallel:
                        if isinstance(self[i, j], chypre.CHypreMat):
                            gcsr = self[i, j]
                            cp = self[i, j].GetColPartArray()
                            rp = self[i, j].GetRowPartArray()
                            s = self[i, j].shape
                            # if (cp == rp).all() and s[0] == s[1]:
                            if gcsr[0] is not None:
                                csr = ToScipyCoo(gcsr[0]).tocsr()
                                gcsr[0] = ToHypreParCSR(csr, col_starts=cp)
                            if gcsr[1] is not None:
                                csr = ToScipyCoo(gcsr[1]).tocsr()
                                gcsr[1] = ToHypreParCSR(csr, col_starts=cp)
                                gcsrm = ToHypreParCSR(-csr, col_starts=cp)
                            dprint2(i, j, s, rp, cp)
                        else:
                            assert False, "unsupported block element " + \
                                str(type(self[i, j]))
                    else:
                        if isinstance(self[i, j], ScipyCoo):
                            gcsr = self[i, j].get_mfem_sparsemat()
                            if gcsr[1] is not None:
                                gcsrm = mfem.SparseMatrix(gcsr[1])
                                gcsrm *= -1.
                        else:
                            assert False, "unsupported block element " + \
                                type(self[i, j])

                    glcsr.SetBlock(ii, jj, gcsr[0])
                    if self.complex:
                        glcsr.SetBlock(ii + 1, jj + 1, gcsr[0])
                    if gcsr[1] is not None:
                        glcsr.SetBlock(ii + 1, jj, gcsr[1])
                        glcsr.SetBlock(ii, jj + 1, gcsrm)
                jj = jj + 2 if self.complex else jj + 1
            ii = ii + 2 if self.complex else ii + 1

        return glcsr

    def gather_blkvec_merged(self, size_hint=None, symmetric=False):
        '''
        Construct MFEM::BlockVector

        This routine ordered unkonws in the following order
           (Re_FES1_node1, Im_FES1_node1, Re_FES1_node2, Im_FES1_node2, ...)

        '''
        assert self.complex, "this format is complex only"

        roffsets = self.get_local_partitioning_v(size_hint=size_hint)
        roffsets = np.sum(np.diff(roffsets).reshape(-1, 2), 1)
        roffsets = np.hstack([0, np.cumsum(roffsets)])

        dprint1("roffsets(vector)", roffsets)
        offset = mfem.intArray(list(roffsets))

        vec = mfem.BlockVector(offset)
        vec._offsets = offset  # in order to keep it from freed

        data = []
        ii = 0
        jj = 0

        # Here I don't like that I am copying the data between two vectors..
        # But, avoiding this takes the large rearangement of program flow...
        for i in range(self.shape[0]):
            if self[i, 0] is not None:
                if isinstance(self[i, 0], chypre.CHypreVec):
                    rr = self[i, 0][0].GetDataArray()
                    if self[i, 0][1] is not None:
                        if symmetric:
                            ii = -self[i, 0][1].GetDataArray()
                        else:
                            ii = self[i, 0][1].GetDataArray()
                    else:
                        ii = rr * 0
                    vv = np.hstack([rr, ii])
                    vec.GetBlock(i).Assign(vv)
                elif isinstance(self[i, 0], ScipyCoo):
                    arr = np.atleast_1d(self[i, 0].toarray().squeeze())
                    if symmetric:
                        arr = np.hstack([np.real(arr), -np.imag(arr)])
                    else:
                        arr = np.hstack([np.real(arr), np.imag(arr)])
                    vec.GetBlock(i).Assign(arr)
                else:
                    assert False, "not implemented, " + str(type(self[i, 0]))
            else:
                vec.GetBlock(i).Assign(0.0)
        return vec

    def get_global_blkmat_merged(self, symmetric=False):
        '''
        This routine ordered unkonws in the following order
           Re FFE1, Im FES1, ReFES2, Im FES2, ...
        matrix FES1, FES2...
        '''
        assert self.complex, "this format is complex only"

        roffsets, coffsets = self.get_local_partitioning()
        roffsets = np.sum(np.diff(roffsets).reshape(-1, 2), 1)
        coffsets = np.sum(np.diff(coffsets).reshape(-1, 2), 1)
        roffsets = np.hstack([0, np.cumsum(roffsets)])
        coffsets = np.hstack([0, np.cumsum(coffsets)])
        dprint1("Generating MFEM BlockMatrix: shape = " +
                str((len(roffsets) - 1, len(coffsets) - 1)))
        dprint1(
            "Generating MFEM BlockMatrix: roffset/coffset = ",
            roffsets,
            coffsets)

        ro = mfem.intArray(list(roffsets))
        co = mfem.intArray(list(coffsets))
        glcsr = mfem.BlockOperator(ro, co)

        ii = 0

        for j in range(self.shape[1]):
            jfirst = True
            for i in range(self.shape[0]):
                if self[i, j] is not None:
                    if use_parallel:
                        if isinstance(self[i, j], chypre.CHypreMat):
                            gcsr = self[i, j]
                            cp = self[i, j].GetColPartArray()
                            rp = self[i, j].GetRowPartArray()
                            rsize_local = rp[1] - rp[0]

                            if jfirst:
                                csize_local = cp[1] - cp[0]
                                csize = allgather(csize_local)
                                cstarts = np.hstack([0, np.cumsum(csize)])
                                jfirst = False
                            s = self[i, j].shape
                            # if (cp == rp).all() and s[0] == s[1]:

                            '''
                            if gcsr[0] is not None:
                                csc = ToScipyCoo(gcsr[0]).tocsc()
                                csc1 = [csc[:,cstarts[k]:cstarts[k+1]] for k in range(len(cstarts)-1)]
                                if symmetric:
                                   csc1m = [-csc[:,cstarts[k]:cstarts[k+1]] for k in range(len(cstarts)-1)]
                            else:
                                csc1 = [None]*len(csize)
                                csc1m = [None]*len(csize)

                            if gcsr[1] is not None:
                                csc   = ToScipyCoo(gcsr[1]).tocsc()
                                if not symmetric:
                                    csc2  = [ csc[:,cstarts[k]:cstarts[k+1]] for k in range(len(cstarts)-1)]
                                csc2m = [-csc[:,cstarts[k]:cstarts[k+1]] for k in range(len(cstarts)-1)]
                            else:
                                csc2  = [None]*len(csize)
                                csc2m = [None]*len(csize)

                            if symmetric:
                                csr = scipy.sparse.bmat([sum(zip(csc1, csc2m), ()),
                                                         sum(zip(csc2m, csc1m),  ())]).tocsr()
                            else:
                                csr = scipy.sparse.bmat([sum(zip(csc1, csc2m), ()),
                                                         sum(zip(csc2, csc1),  ())]).tocsr()

                            cp2 = [(cp*2)[0], (cp*2)[1], np.sum(csize)*2]
                            #gcsr[1] = ToHypreParCSR(csr, col_starts =cp)
                            gcsr  = ToHypreParCSR(csr, col_starts =cp2)
                            '''
                            if gcsr[0] is None:
                                csr = scipy.sparse.csr_matrix(
                                    (rsize_local, cstarts[-1]), dtype=np.float64)
                                cp3 = [cp[0], cp[1], cstarts[-1]]
                                gcsa = ToHypreParCSR(csr, col_starts=cp3)
                            else:
                                gcsa = gcsr[0]

                            if gcsr[1] is None:
                                csr = scipy.sparse.csr_matrix(
                                    (rsize_local, cstarts[-1]), dtype=np.float64)
                                cp3 = [cp[0], cp[1], cstarts[-1]]
                                gcsb = ToHypreParCSR(csr, col_starts=cp3)
                            else:
                                gcsb = gcsr[1]

                        else:
                            assert False, "unsupported block element " + \
                                str(type(self[i, j]))
                    else:
                        if isinstance(self[i, j], ScipyCoo):
                            '''
                            if symmetric:
                                tmp  = scipy.sparse.bmat([[ self[i, j].real, -self[i, j].imag],
                                                          [-self[i, j].imag, -self[i, j].real]])
                            else:
                                tmp  = scipy.sparse.bmat([[self[i, j].real, -self[i, j].imag],
                                                          [self[i, j].imag,  self[i, j].real]])
                            '''
                            csra = self[i, j].real.tocsr()
                            csrb = self[i, j].imag.tocsr()
                            csra.eliminate_zeros()
                            csrb.eliminate_zeros()
                            gcsa = mfem.SparseMatrix(csra)
                            gcsb = mfem.SparseMatrix(csrb)
                        else:
                            assert False, "unsupported block element " + \
                                type(self[i, j])

                    Hermitian = False if symmetric else True
                    gcsr = mfem.ComplexOperator(
                        gcsa, gcsb, False, False, Hermitian)

                    gcsr._real_operator = gcsa
                    gcsr._imag_operator = gcsb

                    glcsr.SetBlock(i, j, gcsr)

        return glcsr



from .solver_model import LinearSolverModel, LinearSolver
import numpy as np

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('GMRESModel')

from petram.mfem_config import use_parallel
if use_parallel:
   from petram.helper.mpi_recipes import *
   from mfem.common.parcsr_extra import *
   import mfem.par as mfem
   default_kind = 'hypre'
   
   from mpi4py import MPI                               
   num_proc = MPI.COMM_WORLD.size
   myid     = MPI.COMM_WORLD.rank
   smyid = '{:0>6d}'.format(myid)
   from mfem.common.mpi_debug import nicePrint
   
else:
   import mfem.ser as mfem
   default_kind = 'scipy'

from petram.solver.mumps_model import MUMPSPreconditioner   
SparseSmootherCls = {"Jacobi": (mfem.DSmoother, 0),
                     "l1Jacobi": (mfem.DSmoother, 1),
                     "lumpedJacobi": (mfem.DSmoother, 2),
                     "GS": (mfem.GSSmoother, 0),
                     "forwardGS": (mfem.GSSmoother, 1),
                     "backwardGS": (mfem.GSSmoother, 2),
                     "MUMPS": (MUMPSPreconditioner, None),}                                          

class GMRES(LinearSolverModel):
    has_2nd_panel = False
    accept_complex = False
    always_new_panel = False    
    def init_solver(self):
        pass
    
    def panel1_param(self):
        from petram.pi.widget_smoother import WidgetSmoother       
        return [["log_level",   self.log_level,  400, {}],
                ["max  iter.",  self.maxiter,  400, {}],
                ["rel. tol",    self.reltol,  300,  {}],
                ["abs. tol.",   self.abstol,  300, {}],
                ["restart(kdim)", self.kdim,     400, {}],
                [None, None, 99, {"UI":WidgetSmoother, "span":(1,2)}],
                ["write matrix",  self.write_mat,   3, {"text":""}],]     
    
    def get_panel1_value(self):
        # this will set _mat_weight
        from petram.solver.solver_model import SolveStep
        p = self.parent
        while not isinstance(p, SolveStep):
            p = p.parent
            if p is None:
                assert False, "GMRES is not under SolveStep"
        num_matrix = p.get_num_matrix(self.get_phys())
        
        all_dep_vars = self.root()['Phys'].all_dependent_vars(num_matrix,
                                                              self.get_phys(),
                                                              self.get_phys_range())

        
        prec = [x for x in  self.preconditioners if x[0] in all_dep_vars]        
        names = [x[0] for x in  prec]
        for n in all_dep_vars:
           if not n in names:
              prec.append((n, ['None', 'None']))
        self.preconditioners = prec
        
        return (int(self.log_level), int(self.maxiter),
                self.reltol, self.abstol, int(self.kdim),
                self.preconditioners, self.write_mat)
    
    def import_panel1_value(self, v):
        self.log_level = int(v[0])
        self.maxiter = int(v[1])
        self.reltol = v[2]
        self.abstol = v[3]
        self.kdim = int(v[4])
        self.preconditioners = v[5]
        self.write_mat = int(v[6])
        
    def attribute_set(self, v):
        v = super(GMRES, self).attribute_set(v)
        v['log_level'] = 0
        v['maxiter'] = 200
        v['reltol']  = 1e-7
        v['abstol'] = 1e-7
        v['kdim'] =   50
        v['printit'] = 1
        v['preconditioner'] = ''
        v['preconditioners'] = []        
        v['write_mat'] = False        
        return v

    def does_linearsolver_choose_linearsystem_type(self):
        return True
     
    def verify_setting(self):
        if not self.parent.assemble_real:
            for phys in self.get_phys():
                if phys.is_complex():
                    return False, "GMRES does not support complex.", "A complex problem must be converted to a real value problem"
        return True, "", ""

    def linear_system_type(self, assemble_real, phys_complex):
        #if not phys_complex: return 'block'
        return 'blk_interleave'
        #return None

    def real_to_complex(self, solall, M):
        if use_parallel:
           from mpi4py import MPI
           myid     = MPI.COMM_WORLD.rank


           offset = M.RowOffsets().ToList()
           of = [np.sum(MPI.COMM_WORLD.allgather(np.int32(o))) for o in offset]
           if myid != 0: return           

        else:
           offset = M.RowOffsets()
           of = offset.ToList()
           
        rows = M.NumRowBlocks()
        s = solall.shape
        nb = rows/2
        i = 0
        pt = 0
        result = np.zeros((s[0]/2, s[1]), dtype='complex')
        for j in range(nb):
           l = of[i+1]-of[i]
           result[pt:pt+l,:] = (solall[of[i]:of[i+1],:]
                             +  1j*solall[of[i+1]:of[i+2],:])
           i = i+2
           pt = pt + l

        return result

    def allocate_solver(self, datatype='D', engine=None):
        solver = GMRESSolver(self, engine, int(self.maxiter),
                             self.abstol, self.reltol, int(self.kdim))
        #solver.AllocSolver(datatype)
        return solver
     
    def get_possible_child(self):
        '''
        Preconditioners....
        '''
        choice = []
        try:
            from petram.solver.mumps_model import MUMPS
            choice.append(MUMPS)
        except ImportError:
            pass
        return choice
     

class GMRESSolver(LinearSolver):
    is_iterative = True
    def __init__(self, gui, engine, maxiter, abstol, reltol, kdim):
        self.maxiter = maxiter
        self.abstol = abstol
        self.reltol = reltol
        self.kdim = kdim
        LinearSolver.__init__(self, gui, engine)

    def SetOperator(self, opr, dist=False, name = None):
        self.Aname = name
        self.A = opr                     
                             
    def Mult(self, b, x=None, case_base=0):
        if use_parallel:
            return self.solve_parallel(self.A, b, x)
        else:
            return self.solve_serial(self.A, b, x)
                             
    def solve_parallel(self, A, b, x=None):
        from mpi4py import MPI
        myid     = MPI.COMM_WORLD.rank
        nproc    = MPI.COMM_WORLD.size
        from petram.helper.mpi_recipes import gather_vector
        
        def get_block(Op, i, j):
            try:
               return Op._linked_op[(i,j)]
            except KeyError:
               return None

        offset = A.RowOffsets()
        rows = A.NumRowBlocks()
        cols = A.NumColBlocks()
        
        if self.gui.write_mat:
           for i in range(cols):
              for j in range(rows):
                 m = get_block(A, i, j)
                 if m is None: continue
                 m.Print('matrix_'+str(i)+'_'+str(j))
           for i, bb  in enumerate(b):
              for j in range(rows):
                 v = bb.GetBlock(j)
                 v.Print('rhs_'+str(i)+'_'+str(j)+'.'+smyid)
           if x is not None:
              for j in range(rows):
                 xx = x.GetBlock(j)
                 xx.Print('x_'+str(i)+'_'+str(j)+'.'+smyid)
              

        M = mfem.BlockDiagonalPreconditioner(offset)
        
        prcs = dict(self.gui.preconditioners)
        name = self.Aname
        assert not self.gui.parent.is_complex(), "can not solve complex"
        if self.gui.parent.is_converted_from_complex():
           name = sum([[n, n] for n in name], [])
           
        for k, n in enumerate(name):
           prc = prcs[n][1]
           if prc == "None": continue
           name = "".join([tmp for tmp in prc if not tmp.isdigit()])
           
           A0 = get_block(A, k, k)
           if A0 is None and not name.startswith('schur'): continue

           if hasattr(mfem.HypreSmoother, prc):
               invA0 = mfem.HypreSmoother(A0)
               invA0.SetType(getattr(mfem.HypreSmoother, prc))
           elif prc == 'ams':
               depvar = self.engine.r_dep_vars[k]
               dprint1("setting up AMS for ", depvar)
               prec_fespace = self.engine.fespaces[depvar]
               invA0 = mfem.HypreAMS(A0, prec_fespace)
               invA0.SetSingularProblem()
           elif name == 'MUMPS':
               cls = SparseSmootherCls[name][0]
               invA0 = cls(A0, gui=self.gui[prc], engine=self.engine)
           elif name.startswith('schur'):
               args = name.split("(")[-1].split(")")[0].split(",")
               dprint1("setting up schur for ", args)
               if len(args) > 1:
                   assert False, "not yet supported"
               for arg in args:
                    r1 = self.engine.dep_var_offset(arg.strip())
                    c1 = self.engine.r_dep_var_offset(arg.strip())                    
                    B  = get_block(A, k, c1)
                    Bt = get_block(A, r1, k).Transpose()
                    Bt = Bt.Transpose()
                    B0 = get_block(A, r1, c1)
                    Md = mfem.HypreParVector(MPI.COMM_WORLD,
                                             B0.GetGlobalNumRows(),
                                             B0.GetColStarts())
                    B0.GetDiag(Md)
                    Bt.InvScaleRows(Md)
                    S = mfem.ParMult(B, Bt)
                    invA0 = mfem.HypreBoomerAMG(S)
                    invA0.iterative_mode = False
           else:
               cls = SparseSmootherCls[name][0]
               invA0 = cls(A0, gui=self.gui[prc])
               
           invA0.iterative_mode = False
           M.SetDiagonalBlock(k, invA0)
           
        '''
        We should support Shur complement type preconditioner
        if offset.Size() > 2:
            B =  get_block(A, 1, 0)
            MinvBt = get_block(A, 0, 1)
            #Md = mfem.HypreParVector(MPI.COMM_WORLD,
            #                        A0.GetGlobalNumRows(),
            #                        A0.GetRowStarts())
            Md = mfem.Vector()
            A0.GetDiag(Md)
            MinvBt.InvScaleRows(Md)
            S = mfem.ParMult(B, MinvBt)
            invS = mfem.HypreBoomerAMG(S)
            invS.iterative_mode = False
            M.SetDiagonalBlock(1, invS)
        '''
        maxiter = int(self.maxiter)
        atol = self.abstol
        rtol = self.reltol
        kdim = int(self.kdim)
        printit = 1

        sol = []

        solver = mfem.GMRESSolver(MPI.COMM_WORLD)
        solver.SetKDim(kdim)

        
        #solver = mfem.MINRESSolver(MPI.COMM_WORLD)
        #solver.SetOperator(A)        

        #solver = mfem.CGSolver(MPI.COMM_WORLD)
        solver.SetOperator(A)        
        solver.SetAbsTol(atol)
        solver.SetRelTol(rtol)
        solver.SetMaxIter(maxiter)
        solver.SetPreconditioner(M)
        solver.SetPrintLevel(1)

        # solve the problem and gather solution to head node...
        # may not be the best approach

        for bb in b:
           rows = MPI.COMM_WORLD.allgather(np.int32(bb.Size()))
           rowstarts = np.hstack((0, np.cumsum(rows)))
           dprint1("rowstarts/offser",rowstarts, offset.ToList())
           if x is None:           
              xx = mfem.BlockVector(offset)
              xx.Assign(0.0)
           else:
              xx = x
              #for j in range(cols):
              #   dprint1(x.GetBlock(j).Size())
              #   dprint1(x.GetBlock(j).GetDataArray())
              #assert False, "must implement this"
           solver.Mult(bb, xx)
           s = []
           for i in range(offset.Size()-1):
               v = xx.GetBlock(i).GetDataArray()
               vv = gather_vector(v)
               if myid == 0:
                   s.append(vv)
               else:
                   pass
           if myid == 0:               
               sol.append(np.hstack(s))
        if myid == 0:                              
            sol = np.transpose(np.vstack(sol))
            return sol
        else:
            return None
        
    def solve_serial(self, A, b, x=None):

        def get_block(Op, i, j):
            try:
               return Op._linked_op[(i,j)]
            except KeyError:
               return None
                      
        offset = A.RowOffsets()
        rows = A.NumRowBlocks()
        cols = A.NumColBlocks()
        if self.gui.write_mat:
           for i in range(cols):
              for j in range(rows):
                 m = get_block(A, i, j)
                 if m is None: continue
                 m.Print('matrix_'+str(i)+'_'+str(j))
           for i, bb  in enumerate(b):
              for j in range(rows):
                 v = bb.GetBlock(j)
                 v.Print('rhs_'+str(i)+'_'+str(j))

        M = mfem.BlockDiagonalPreconditioner(offset)

        prcs = dict(self.gui.preconditioners)
        name = self.Aname
        assert not self.gui.parent.is_complex(), "can not solve complex"
        if self.gui.parent.is_converted_from_complex():
           name = sum([[n, n] for n in name], [])

           
        '''   
        this if block does a generalized version of this...
        M1 = mfem.GSSmoother(get_block(A, 0, 0))
        M1.iterative_mode = False
        M.SetDiagonalBlock(0, M1)
        '''
        for k, n in enumerate(name):
           
           prc = prcs[n][0]
           if prc == "None": continue
           name = "".join([tmp for tmp in prc if not tmp.isdigit()])
           A0 = get_block(A, k, k)
           cls = SparseSmootherCls[name][0]
           arg = SparseSmootherCls[name][1]
           if name == 'MUMPS':
               invA0 = cls(A0, gui=self.gui[prc], engine=self.engine)
           else:
               invA0 = cls(A0, arg)
           invA0.iterative_mode = False
           M.SetDiagonalBlock(k, invA0)

        '''
        We should support Shur complement type preconditioner
        if offset.Size() > 2:
            B =  get_block(A, 1, 0)
            MinvBt = get_block(A, 0, 1)
            Md = mfem.Vector(get_block(A, 0, 0).Height())
            get_block(A, 0, 0).GetDiag(Md)
            for i in range(Md.Size()):
                if Md[i] != 0.:
                    MinvBt.ScaleRow(i, 1/Md[i])
                else:
                    assert False, "diagnal element of matrix is zero"
            S = mfem.Mult(B, MinvBt)
            S.iterative_mode = False
            SS = mfem.DSmoother(S)
            SS.iterative_mode = False            
            M.SetDiagonalBlock(1, SS)
        '''

        '''
        int GMRES(const Operator &A, Vector &x, const Vector &b, Solver &M,
          int &max_iter, int m, double &tol, double atol, int printit)
        '''
        maxiter = int(self.maxiter)
        atol = self.abstol
        rtol = self.reltol
        kdim = int(self.kdim)
        printit = 1

        sol = []

        solver = mfem.GMRESSolver()
        solver.SetKDim(kdim)
        #solver = mfem.MINRESSolver()
        solver.SetAbsTol(atol)
        solver.SetRelTol(rtol)
        solver.SetMaxIter(maxiter)

        solver.SetOperator(A)
        solver.SetPreconditioner(M)
        solver.SetPrintLevel(1)

        for bb in b:
           if x is None:           
              xx = mfem.Vector(bb.Size())
              xx.Assign(0.0)
           else:
              xx = x
              #for j in range(cols):
              #   print x.GetBlock(j).Size()
              #   print x.GetBlock(j).GetDataArray()                 
              #assert False, "must implement this"
           solver.Mult(bb, xx)
           sol.append(xx.GetDataArray().copy())
        sol = np.transpose(np.vstack(sol))
        return sol
    

         
        
